"""
WordNet model

Author: John McCrae <john@mccr.ae> for original code
Author: Bernard Bou <1313ou@gmail.com> for rewrite and revamp
"""

#  Copyright (c) 2024.
#  Creative Commons 4 for original code
#  GPL3 for rewrite

from enum import StrEnum, Enum
from typing import Any, Optional, Tuple, List, Dict, Set, Generator


class Entry:
    """The lexical entry consists of a single word"""

    def __init__(self, lemma, pos, discriminant) -> None:
        self.lemma: str = lemma
        self.pos: str = pos
        self.discriminant: str | None = discriminant
        self.forms: List[str] = []
        self.pronunciations: List[Pronunciation] = []
        self.senses: List[Sense] = []

    def __str__(self) -> str:
        return f'{self.lemma},{self.pos}'

    def __repr__(self) -> str:
        return str(self.key)

    @property
    def key(self) -> Tuple[str, str, str | None]:
        return self.lemma, self.pos, self.discriminant


class Sense:
    """ The sense links an entry to a synset """

    def __init__(self, senseid, entry, synsetid, adjposition=None) -> None:
        self.id: str = senseid
        self.entry: Entry = entry
        self.synsetid: str = synsetid
        self.resolved_synset: Optional[Synset] = None
        self.adjposition: Optional[str] = adjposition
        self.examples: List[str] = []
        self.verbframeids: Optional[List[str]] = None
        self.relations: List[Sense.Relation] = []

    def __str__(self) -> str:
        return f'{self.id}'

    def __repr__(self) -> str:
        return f'{self.id} @{self.synsetid}'

    def __getstate__(self) -> Dict[str, Any]:
        state = self.__dict__.copy()
        if 'resolved_synset' in state:
            del state['resolved_synset']  # exclude from being pickled
        return state

    def __setstate__(self, state) -> None:
        self.__dict__.update(state)
        self.resolved_synset = None  # restore o a default or None value

    class Relation:
        """ Lexical relation (sense to sense)"""

        class Type(StrEnum):
            ANTONYM: str = 'antonym'
            ALSO: str = 'also'
            PARTICIPLE: str = 'participle'
            PERTAINYM: str = 'pertainym'
            DERIVATION: str = 'derivation'
            DOMAIN_TOPIC: str = 'domain_topic'
            HAS_DOMAIN_TOPIC: str = 'has_domain_topic'
            DOMAIN_REGION: str = 'domain_region'
            HAS_DOMAIN_REGION: str = 'has_domain_region'
            EXEMPLIFIES: str = 'exemplifies'
            IS_EXEMPLIFIED_BY: str = 'is_exemplified_by'
            SIMILAR: str = 'similar'
            COLLOCATION: str = 'collocation'
            OTHER: str = 'other'

        inverses: Dict[Type, Type] = {
            Type.DOMAIN_REGION: Type.HAS_DOMAIN_REGION,
            Type.HAS_DOMAIN_REGION: Type.DOMAIN_REGION,
            Type.DOMAIN_TOPIC: Type.HAS_DOMAIN_TOPIC,
            Type.HAS_DOMAIN_TOPIC: Type.DOMAIN_TOPIC,
            Type.EXEMPLIFIES: Type.IS_EXEMPLIFIED_BY,
            Type.IS_EXEMPLIFIED_BY: Type.EXEMPLIFIES,
            Type.ANTONYM: Type.ANTONYM,
            Type.SIMILAR: Type.SIMILAR,
            Type.COLLOCATION: Type.COLLOCATION,
            Type.ALSO: Type.ALSO,
            Type.DERIVATION: Type.DERIVATION,
        }

        class OtherType(StrEnum):
            AGENT: str = 'agent'
            MATERIAL: str = 'material'
            EVENT: str = 'event'
            INSTRUMENT: str = 'instrument'
            LOCATION: str = 'location'
            BY_MEANS_OF: str = 'by_means_of'
            UNDERGOER: str = 'undergoer'
            PROPERTY: str = 'property'
            RESULT: str = 'result'
            STATE: str = 'state'
            USES: str = 'uses'
            DESTINATION: str = 'destination'
            BODY_PART: str = 'body_part'
            VEHICLE: str = 'vehicle'

        def __init__(self, target, relation_type, other_type=False) -> None:
            self.target: str = target
            self.resolved_target: Optional[Sense] = None
            self.relation_type: str = relation_type
            self.other_type: bool = other_type

        def __str__(self) -> str:
            return f'-{self.relation_type}-> {self.target}'

        def __repr__(self) -> str:
            return f'{self.relation_type}: {self.target}'

        def __getstate__(self) -> dict[str, Any]:
            state = self.__dict__.copy()
            if 'resolved_target' in state:
                del state['resolved_target']  # exclude from being pickled
            return state

        def __setstate__(self, state) -> None:
            self.__dict__.update(state)
            self.resolved_target = None  # restore o a default or None value


class Synset:
    """ Synset, a collection of members that share a common meaning """

    def __init__(self, synsetid, pos, members, lex_name) -> None:
        self.id: str = synsetid
        self.pos: str = PartOfSpeech(pos).value
        self.members: List[str] = members
        self.resolved_members: Optional[List[Entry]] = None
        self.lex_name: str = lex_name
        self.definitions: List[str] = []
        self.examples: List[str | Example] = []
        self.usages: List[str] = []
        self.ili_definition: Optional[str] = None
        self.source: Optional[str] = None
        self.wikidata: Optional[str] = None
        self.ili: Optional[str] = None
        self.relations: List[Synset.Relation] = []

    def __str__(self) -> str:
        return f'{self.id}'

    def __repr__(self) -> str:
        return f'{self.id} [{' '.join(self.members)}]'

    def __getstate__(self) -> Dict[str, Any]:
        state = self.__dict__.copy()
        if 'resolved_members' in state:
            del state['resolved_members']  # exclude from being pickled
        return state

    def __setstate__(self, state) -> None:
        self.__dict__.update(state)
        self.resolved_members = None  # restore o a default or None value

    class Relation:
        """ Semantic relation (synset to synset)"""

        class Type(StrEnum):
            AGENT: str = 'agent'
            ALSO: str = 'also'
            ATTRIBUTE: str = 'attribute'
            BE_IN_STATE: str = 'be_in_state'
            CAUSES: str = 'causes'
            CLASSIFIED_BY: str = 'classified_by'
            CLASSIFIES: str = 'classifies'
            CO_AGENT_INSTRUMENT: str = 'co_agent_instrument'
            CO_AGENT_PATIENT: str = 'co_agent_patient'
            CO_AGENT_RESULT: str = 'co_agent_result'
            CO_INSTRUMENT_AGENT: str = 'co_instrument_agent'
            CO_INSTRUMENT_PATIENT: str = 'co_instrument_patient'
            CO_INSTRUMENT_RESULT: str = 'co_instrument_result'
            CO_PATIENT_AGENT: str = 'co_patient_agent'
            CO_PATIENT_INSTRUMENT: str = 'co_patient_instrument'
            CO_RESULT_AGENT: str = 'co_result_agent'
            CO_RESULT_INSTRUMENT: str = 'co_result_instrument'
            CO_ROLE: str = 'co_role'
            DIRECTION: str = 'direction'
            DOMAIN_REGION: str = 'domain_region'
            DOMAIN_TOPIC: str = 'domain_topic'
            EXEMPLIFIES: str = 'exemplifies'
            ENTAILS: str = 'entails'
            EQ_SYNONYM: str = 'eq_synonym'
            HAS_DOMAIN_REGION: str = 'has_domain_region'
            HAS_DOMAIN_TOPIC: str = 'has_domain_topic'
            IS_EXEMPLIFIED_BY: str = 'is_exemplified_by'
            HOLO_LOCATION: str = 'holo_location'
            HOLO_MEMBER: str = 'holo_member'
            HOLO_PART: str = 'holo_part'
            HOLO_PORTION: str = 'holo_portion'
            HOLO_SUBSTANCE: str = 'holo_substance'
            HOLONYM: str = 'holonym'
            HYPERNYM: str = 'hypernym'
            HYPONYM: str = 'hyponym'
            IN_MANNER: str = 'in_manner'
            INSTANCE_HYPERNYM: str = 'instance_hypernym'
            INSTANCE_HYPONYM: str = 'instance_hyponym'
            INSTRUMENT: str = 'instrument'
            INVOLVED: str = 'involved'
            INVOLVED_AGENT: str = 'involved_agent'
            INVOLVED_DIRECTION: str = 'involved_direction'
            INVOLVED_INSTRUMENT: str = 'involved_instrument'
            INVOLVED_LOCATION: str = 'involved_location'
            INVOLVED_PATIENT: str = 'involved_patient'
            INVOLVED_RESULT: str = 'involved_result'
            INVOLVED_SOURCE_DIRECTION: str = 'involved_source_direction'
            INVOLVED_TARGET_DIRECTION: str = 'involved_target_direction'
            IS_CAUSED_BY: str = 'is_caused_by'
            IS_ENTAILED_BY: str = 'is_entailed_by'
            LOCATION: str = 'location'
            MANNER_OF: str = 'manner_of'
            MERO_LOCATION: str = 'mero_location'
            MERO_MEMBER: str = 'mero_member'
            MERO_PART: str = 'mero_part'
            MERO_PORTION: str = 'mero_portion'
            MERO_SUBSTANCE: str = 'mero_substance'
            MERONYM: str = 'meronym'
            SIMILAR: str = 'similar'
            OTHER: str = 'other'
            PATIENT: str = 'patient'
            RESTRICTED_BY: str = 'restricted_by'
            RESTRICTS: str = 'restricts'
            RESULT: str = 'result'
            ROLE: str = 'role'
            SOURCE_DIRECTION: str = 'source_direction'
            STATE_OF: str = 'state_of'
            TARGET_DIRECTION: str = 'target_direction'
            SUBEVENT: str = 'subevent'
            IS_SUBEVENT_OF: str = 'is_subevent_of'
            ANTONYM: str = 'antonym'

        inverses: Dict[Type, Type] = {
            Type.HYPERNYM: Type.HYPONYM,
            Type.HYPONYM: Type.HYPERNYM,
            Type.INSTANCE_HYPERNYM: Type.INSTANCE_HYPONYM,
            Type.INSTANCE_HYPONYM: Type.INSTANCE_HYPERNYM,
            Type.MERONYM: Type.HOLONYM,
            Type.HOLONYM: Type.MERONYM,
            Type.MERO_LOCATION: Type.HOLO_LOCATION,
            Type.HOLO_LOCATION: Type.MERO_LOCATION,
            Type.MERO_MEMBER: Type.HOLO_MEMBER,
            Type.HOLO_MEMBER: Type.MERO_MEMBER,
            Type.MERO_PART: Type.HOLO_PART,
            Type.HOLO_PART: Type.MERO_PART,
            Type.MERO_PORTION: Type.HOLO_PORTION,
            Type.HOLO_PORTION: Type.MERO_PORTION,
            Type.MERO_SUBSTANCE: Type.HOLO_SUBSTANCE,
            Type.HOLO_SUBSTANCE: Type.MERO_SUBSTANCE,
            Type.BE_IN_STATE: Type.STATE_OF,
            Type.STATE_OF: Type.BE_IN_STATE,
            Type.CAUSES: Type.IS_CAUSED_BY,
            Type.IS_CAUSED_BY: Type.CAUSES,
            Type.SUBEVENT: Type.IS_SUBEVENT_OF,
            Type.IS_SUBEVENT_OF: Type.SUBEVENT,
            Type.MANNER_OF: Type.IN_MANNER,
            Type.IN_MANNER: Type.MANNER_OF,
            Type.RESTRICTS: Type.RESTRICTED_BY,
            Type.RESTRICTED_BY: Type.RESTRICTS,
            Type.CLASSIFIES: Type.CLASSIFIED_BY,
            Type.CLASSIFIED_BY: Type.CLASSIFIES,
            Type.ENTAILS: Type.IS_ENTAILED_BY,
            Type.IS_ENTAILED_BY: Type.ENTAILS,
            Type.DOMAIN_REGION: Type.HAS_DOMAIN_REGION,
            Type.HAS_DOMAIN_REGION: Type.DOMAIN_REGION,
            Type.DOMAIN_TOPIC: Type.HAS_DOMAIN_TOPIC,
            Type.HAS_DOMAIN_TOPIC: Type.DOMAIN_TOPIC,
            Type.EXEMPLIFIES: Type.IS_EXEMPLIFIED_BY,
            Type.IS_EXEMPLIFIED_BY: Type.EXEMPLIFIES,
            Type.ROLE: Type.INVOLVED,
            Type.INVOLVED: Type.ROLE,
            Type.AGENT: Type.INVOLVED_AGENT,
            Type.INVOLVED_AGENT: Type.AGENT,
            Type.PATIENT: Type.INVOLVED_PATIENT,
            Type.INVOLVED_PATIENT: Type.PATIENT,
            Type.RESULT: Type.INVOLVED_RESULT,
            Type.INVOLVED_RESULT: Type.RESULT,
            Type.INSTRUMENT: Type.INVOLVED_INSTRUMENT,
            Type.INVOLVED_INSTRUMENT: Type.INSTRUMENT,
            Type.LOCATION: Type.INVOLVED_LOCATION,
            Type.INVOLVED_LOCATION: Type.LOCATION,
            Type.DIRECTION: Type.INVOLVED_DIRECTION,
            Type.INVOLVED_DIRECTION: Type.DIRECTION,
            Type.TARGET_DIRECTION: Type.INVOLVED_TARGET_DIRECTION,
            Type.INVOLVED_TARGET_DIRECTION: Type.TARGET_DIRECTION,
            Type.SOURCE_DIRECTION: Type.INVOLVED_SOURCE_DIRECTION,
            Type.INVOLVED_SOURCE_DIRECTION: Type.SOURCE_DIRECTION,
            Type.CO_AGENT_PATIENT: Type.CO_PATIENT_AGENT,
            Type.CO_PATIENT_AGENT: Type.CO_AGENT_PATIENT,
            Type.CO_AGENT_INSTRUMENT: Type.CO_INSTRUMENT_AGENT,
            Type.CO_INSTRUMENT_AGENT: Type.CO_AGENT_INSTRUMENT,
            Type.CO_AGENT_RESULT: Type.CO_RESULT_AGENT,
            Type.CO_RESULT_AGENT: Type.CO_AGENT_RESULT,
            Type.CO_PATIENT_INSTRUMENT: Type.CO_INSTRUMENT_PATIENT,
            Type.CO_INSTRUMENT_PATIENT: Type.CO_PATIENT_INSTRUMENT,
            Type.CO_RESULT_INSTRUMENT: Type.CO_INSTRUMENT_RESULT,
            Type.CO_INSTRUMENT_RESULT: Type.CO_RESULT_INSTRUMENT,
            Type.ANTONYM: Type.ANTONYM,
            Type.EQ_SYNONYM: Type.EQ_SYNONYM,
            Type.SIMILAR: Type.SIMILAR,
            # Type.ALSO: Type.ALSO,
            Type.ATTRIBUTE: Type.ATTRIBUTE,
            Type.CO_ROLE: Type.CO_ROLE
        }

        def __init__(self, target, relation_type) -> None:
            self.target: str = target
            self.resolved_target: Optional[Synset] = None
            self.relation_type: str = relation_type

        def __str__(self) -> str:
            return f'-{self.relation_type}-> {self.target}'

        def __repr__(self) -> str:
            return f'{self.relation_type}: {self.target}'

        def __getstate__(self) -> dict[str, Any]:
            state = self.__dict__.copy()
            if 'resolved_target' in state:
                del state['resolved_target']  # exclude from being pickled
            return state

        def __setstate__(self, state) -> None:
            self.__dict__.update(state)
            self.resolved_target = None  # restore o a default or None value


ignored_symmetric_synset_relations: Set[Synset.Relation.Type] = {
    Synset.Relation.Type.HYPONYM,
    Synset.Relation.Type.INSTANCE_HYPONYM,
    Synset.Relation.Type.HOLONYM,
    Synset.Relation.Type.HOLO_LOCATION,
    Synset.Relation.Type.HOLO_MEMBER,
    Synset.Relation.Type.HOLO_PART,
    Synset.Relation.Type.HOLO_PORTION,
    Synset.Relation.Type.HOLO_SUBSTANCE,
    Synset.Relation.Type.STATE_OF,
    Synset.Relation.Type.IS_CAUSED_BY,
    Synset.Relation.Type.IS_SUBEVENT_OF,
    Synset.Relation.Type.IN_MANNER,
    Synset.Relation.Type.RESTRICTED_BY,
    Synset.Relation.Type.CLASSIFIED_BY,
    Synset.Relation.Type.IS_ENTAILED_BY,
    Synset.Relation.Type.HAS_DOMAIN_REGION,
    Synset.Relation.Type.HAS_DOMAIN_TOPIC,
    Synset.Relation.Type.IS_EXEMPLIFIED_BY,
    Synset.Relation.Type.INVOLVED,
    Synset.Relation.Type.INVOLVED_AGENT,
    Synset.Relation.Type.INVOLVED_PATIENT,
    Synset.Relation.Type.INVOLVED_RESULT,
    Synset.Relation.Type.INVOLVED_INSTRUMENT,
    Synset.Relation.Type.INVOLVED_LOCATION,
    Synset.Relation.Type.INVOLVED_DIRECTION,
    Synset.Relation.Type.INVOLVED_TARGET_DIRECTION,
    Synset.Relation.Type.INVOLVED_SOURCE_DIRECTION,
    Synset.Relation.Type.CO_PATIENT_AGENT,
    Synset.Relation.Type.CO_INSTRUMENT_AGENT,
    Synset.Relation.Type.CO_RESULT_AGENT,
    Synset.Relation.Type.CO_INSTRUMENT_PATIENT,
    Synset.Relation.Type.CO_INSTRUMENT_RESULT
}

ignored_symmetric_sense_relations: Set[Sense.Relation.Type] = {
    Sense.Relation.Type.HAS_DOMAIN_REGION,
    Sense.Relation.Type.HAS_DOMAIN_TOPIC,
    Sense.Relation.Type.IS_EXEMPLIFIED_BY
}


class Pronunciation:
    """ Pronunciation of a lemma"""

    def __init__(self, value, variety=None) -> None:
        self.value: str = value
        self.variety: Optional[str] = variety


class Example:
    """ Sourced example """

    def __init__(self, text, source=None) -> None:
        self.text: str = text
        self.source: Optional[str] = source


class VerbFrame:
    """ Verb frame """

    def __init__(self, fid, verbframe) -> None:
        self.id: str = fid
        self.verbframe: str = verbframe


class PartOfSpeech(Enum):
    """ Parts of speech """

    NOUN: str = 'n'
    VERB: str = 'v'
    ADJECTIVE: str = 'a'
    ADVERB: str = 'r'
    ADJECTIVE_SATELLITE: str = 's'


class WordnetModel:
    """
    The Model/Lexicon contains all the synsets and entries
    """

    def __init__(self, lexicon_id, label, language, email, wnlicense, version, url) -> None:

        # meta data
        self.id: str = lexicon_id
        self.label: str = label
        self.language: str = language
        self.email: str = email
        self.license: str = wnlicense
        self.version: str = version
        self.url: str = url

        # data
        self.entries: List[Entry] = []
        self.synsets: List[Synset] = []
        self.verbframes: List[VerbFrame] = []

        # resolvers
        self.synset_resolver: Dict[str, Synset] = {}
        self.sense_resolver: Dict[str, Sense] = {}
        self.member_resolver: Dict[Tuple[str, str], Entry] = {}  # key is (lemma,synsetid)

    def __str__(self) -> str:
        return f"Wordnet '{self.id}'"

    def info(self) -> str:
        """ Counts """
        return f'{self} has {len(self.entries)} entries, {len(self.synsets)} synsets and {sum(1 for _ in self.senses)} senses'

    def info_relations(self) -> str:
        """ Counts of relations """
        return f'{self} has {sum(1 for _ in self.sense_relations)} sense relations and {sum(1 for _ in self.synset_relations)} synset relations'

    @property
    def entry_resolver(self) -> Dict[Tuple[str, str, str | None], Entry]:
        return {e.key: e for e in self.entries}

    @property
    def senses(self) -> Generator[Sense, None, None]:
        """ Senses generator """
        for e in self.entries:
            for s in e.senses:
                yield s

    @property
    def sense_relations(self) -> Generator[Sense.Relation, None, None]:
        """ Sense relations generator """
        for s in self.senses:
            for r in s.relations:
                yield r

    @property
    def synset_relations(self) -> Generator[Synset.Relation, None, None]:
        """ Synset relations generator """
        for ss in self.synsets:
            for r in ss.relations:
                yield r

    @property
    def verbframe_resolver(self) -> Dict[str, str]:
        """ Verb frame resolver factory from id """
        return {f.id: f.verbframe for f in self.verbframes}

    def extend(self) -> None:
        """
        Extend to include inverse relations can be added here
        """
        for ss in self.synsets:
            self.extend_synset_relations(ss)
        for s in self.senses:
            self.extend_sense_relations(s)

    def extend_sense_relations(self, sense: Sense) -> None:
        """
        Add inverse sense relations as needed
        :param sense: sense to extend
        """
        for r in sense.relations:
            if not r.other_type:
                t = Sense.Relation.Type(r.relation_type)
                if t in Sense.Relation.inverses:
                    inv_t = Sense.Relation.inverses[t]
                    if inv_t != t and t not in ignored_symmetric_sense_relations:
                        target_sense = self.sense_resolver[r.target]
                        if not target_sense:
                            raise ValueError(f'Unresolved target {r.target} in relation of type {t} in sense {sense.id}')
                        if not any(r2 for r2 in target_sense.relations if
                                   r2.target == sense.id and not r2.other_type and Sense.Relation.Type(r2.relation_type) == inv_t):
                            target_sense.relations.append(Sense.Relation(sense.id, inv_t.value))

    def extend_synset_relations(self, synset: Synset) -> None:
        """
        Add inverse synset relations as needed
        :param synset: synset to extend
        """
        for r in synset.relations:
            t = Synset.Relation.Type(r.relation_type)
            if t in Synset.Relation.inverses:
                inv_t = Synset.Relation.inverses[t]
                if inv_t != t and t not in ignored_symmetric_synset_relations:
                    target_synset = self.synset_resolver[r.target]
                    if not target_synset:
                        raise ValueError(f'Unresolved target {r.target} in relation of type {t} in synset {synset.id}')
                    if not any(r2 for r2 in target_synset.relations if r2.target == synset.id and Synset.Relation.Type(r2.relation_type) == inv_t):
                        target_synset.relations.append(Synset.Relation(synset.id, inv_t.value))

    def resolve(self) -> None:
        """
        Resolve model internal cross-references.
        Side effect is computation of resolved_* fields.
        :raises: ValueError when the resolvers are not available
        :raises: KeyError when the resolvers can't resolve the keys
        """
        # sanity check
        if not self.synset_resolver:
            raise ValueError(f'{self} has no synset resolver')
        if not self.sense_resolver:
            raise ValueError(f'{self} has no sense resolver')
        if not self.member_resolver:
            raise ValueError(f'{self} has no member resolver')

        for s in self.senses:
            # resolve synset reference in sense
            s.resolved_synset = self.synset_resolver[s.synsetid]

            for r in s.relations:
                # resolve relation target reference in sense relation
                r.resolved_target = self.sense_resolver[r.target]

        for ss in self.synsets:
            # resolve member references in synset
            ss.resolved_members = [self.member_resolver[(m, ss.id)] for m in ss.members]

            for r in ss.relations:
                # resolve relation target reference in synset relation
                r.resolved_target = self.synset_resolver[r.target]

    def stale(self) -> None:
        """
        Stale all model internal cross-references.
        """
        for s in self.senses:
            s.resolved_synset = None
            for r in s.relations:
                r.resolved_target = None
        for ss in self.synsets:
            ss.resolved_members = None
            for r in ss.relations:
                r.resolved_target = None



"""
WordNet XML ID well-formedness abstract tests
Author: John McCrae <john@mccr.ae> for original code
Author: Bernard Bou <1313ou@gmail.com> for rewrite and revamp
"""
#  Copyright (c) 2024.
#  Creative Commons 4 for original code
#  GPL3 for rewrite

import sys
import unittest
from typing import Tuple

from oewn_xml.wordnet_xml import dash_factory, is_valid_xml_id, is_valid_xml_id_char, escape_lemma, unescape_lemma, escape_sensekey, unescape_sensekey


class WordnetXMLTestCase(unittest.TestCase):
    def test_valid_char(self) -> None:
        for c in ('.', '·', ':', '_', '-'):
            self.assertTrue(is_valid_xml_id_char(c), f'Valid {c} {c!r} x{ord(c):X}')
        for c in (',', ';', '%', '!', '?', '*', '/', '|', '\\', '`'):
            self.assertFalse(is_valid_xml_id_char(c), f'Not valid {c} {c!r} x{ord(c):X}')

    def test_valid_ids(self) -> None:
        for s in (
                'a.b',
                'a·b',
                'a:b',
                'a_b',
                'a-b'
        ):
            self.assertTrue(is_valid_xml_id(s), f'Valid {s}')

        for s in (
                'a,b',
                'a;b',
                'a!b',
                'a?b',
                'a*b',
                'a/b',
                'a|b',
                'a\\b',
                'a`b'
        ):
            self.assertFalse(is_valid_xml_id(s), f'Not valid {s}')

    def test_valid_escape_keys(self) -> None:
        d = dash_factory.base_char_escapes
        for c in d:
            self.assertFalse(is_valid_xml_id_char(c), f'Not valid {c} {c!r} x{ord(c):X}')

    def test_valid_escape_unescape_lemmas(self) -> None:
        for l in some_lemmas:
            r = lemma_escape_unescape(l)
            assert r
            l1, l2 = r
            print(f'{l} -> {l1} -> {l2}')
            self.assertTrue(is_valid_xml_id(l1), f'Illegal XML ID "{l1}"')
            self.assertEqual(l, l2, f'Reversing failed {l} {l2}')

    def test_valid_escape_unescape_sensekeys(self) -> None:
        for l in some_lemmas:
            sk = make_yaml_sense_key(l, 1, 2, 3, None, None)
            # sk = 'a_b_c%1:02:03::'
            r = sensekey_escape_unescape(sk)
            assert r
            sk1, sk2 = r
            print(f'{sk} -> {sk1} -> {sk2}')
            self.assertTrue(is_valid_xml_id(sk1), f'Illegal XML ID "{sk1}"')
            self.assertEqual(sk, sk2, f'Reversing failed {sk} {sk2}')


# UTILS

def lemma_escape_unescape(l: str) -> Tuple[str, str] | None:
    try:
        l1 = escape_lemma(l)
        l2 = unescape_lemma(l1)
        return l1, l2
    except ValueError as er:
        print(f'{l} -> LEMMA ERROR {er}', file=sys.stderr)


def sensekey_escape_unescape(sk: str) -> Tuple[str, str] | None:
    try:
        sk1 = escape_sensekey(sk)
        sk2 = unescape_sensekey(sk1)
        return sk1, sk2
    except Exception as er:
        print(f'{sk} -> SENSEKEY ERROR {er}', file=sys.stderr)


def make_yaml_sense_key(lemma: str, ss_type: int, lex_filenum: int, lexid: int, head: str | None,
                        head_id: int | None) -> str:
    """Calculate the sense key for a sense of an entry"""
    lemma2 = lemma.lower().replace(" ", "_").replace("%", "-percent-")
    return f'{lemma2}%{ss_type}:{lex_filenum:02}:{lexid:02}:{head if head is not None else ''}:{f'{head_id:02}' if head_id is not None else ''}'


some_lemmas: Tuple[str, ...] = (
    'protégé',
    'señor',
    'a.b.c',
    'a-b-c',
    'a_b_c',
    'a b c',
    'a:b:c',  # colon
    'a/b/c',
    'a|b|c',
    'a\\b\\c',
    'a,b,c',
    'a;b;c',
    'a=b=c',
    'a+b+c',
    'a(b)c',
    'a{b}c',
    'a[b]c',
    'a!b!c',
    'a?b?c',
    'a$b$c',
    'a#b#c',
    'a%b%c',  # percent
    'a@b@c',
    'a^b^c',
    'a*b*c',
    "a'b'c",
    'a`b`c',
    'a´b´c',
    'a‘quoted’b',
    'a🠀b🠂c',
)

some_illegal_lemmas: Tuple[str, ...] = (
    'a→b←c',
)

some_lemmas_with_percent_or_colon: Tuple[str, ...] = (
    'a%b%c',
    'a:b:c',
)

if __name__ == '__main__':
    unittest.main()



"""
WordNet compare escape schemes
Author: John McCrae <john@mccr.ae> for original code
Author: Bernard Bou <1313ou@gmail.com> for rewrite and revamp
"""
#  Copyright (c) 2024.
#  Creative Commons 4 for original code
#  GPL3 for rewrite

import random
import unittest
from typing import Tuple, Generator

from oewn_xml.wordnet_xml import dash_factory, is_valid_xml_id
from tests.model import wn
from tests.utils import collect_entries_for_escapes, is_valid_xsd_id, generate_legacy_senseid_from_sensekey, generate_senseid_from_sensekey


def generate_senseids(some_entries, limit=3) -> Generator[Tuple[str, str, str], None, None]:
    """ Produces senseids for comparison"""
    for i, e in enumerate(some_entries):
        for s in e.senses:
            sk, senseid, unesc_sk = generate_senseid_from_sensekey(s.id)
            sk, legacy_senseid, legacy_unesc_sk = generate_legacy_senseid_from_sensekey(s.id)
            if i < limit:
                yield sk, senseid, legacy_senseid


class EscapeSchemesTestCase(unittest.TestCase):
    limit = 5

    def test_compare_escape_schemes_on_random_samples(self) -> None:
        print('\nCOMPARE ESCAPE SCHEMES (RANDOM SELECTION)')
        some_entries = random.sample(wn.entries, min(len(wn.entries), 5))
        for sk, senseid, legacy_senseid in generate_senseids(some_entries, self.limit):
            print(f'\t{sk}')
            print(f'\t\t--now--> {senseid}')
            print(f'\t\t--leg--> {legacy_senseid}')

    def test_compare_escape_schemes_escapables_on_escape_cases(self) -> None:
        print('\nCOMPARE ESCAPE SCHEMES')
        r = collect_entries_for_escapes(wn.entries, dash_factory.char_escapes_for_sk)
        for k in r:
            v = r[k]
            if v:
                print(f'{k}')
                some_entries = random.sample(r[k], min(len(r[k]), 5))
                for sk, senseid, legacy_senseid in generate_senseids(some_entries, self.limit):
                    print(f'\t{sk}')
                    print(f'\t\t--now--> {senseid}')
                    print(f'\t\t--leg--> {legacy_senseid}')

    def test_validity_on_escapables(self) -> None:
        print('\nTEST ESCAPE SCHEMES')
        r = collect_entries_for_escapes(wn.entries, dash_factory.char_escapes_for_sk)
        for k in r:
            v = r[k]
            if v:
                print(f'{k}')
                some_entries = random.sample(r[k], min(len(r[k]), 5))
                for sk, senseid, legacy_senseid in generate_senseids(some_entries, self.limit):
                    print(f'\t{sk} --senseid--> {senseid}')
                    self.assertTrue(is_valid_xml_id(senseid), f'{sk} --senseid--> {senseid}')
                    self.assertTrue(is_valid_xsd_id(senseid), f'{sk} --senseid--> {senseid}')
                    print(f'\t{sk} --legacy senseid--> {legacy_senseid}')
                    self.assertTrue(is_valid_xml_id(legacy_senseid), f'{sk}--legacy senseid--> {legacy_senseid}')
                    self.assertTrue(is_valid_xsd_id(legacy_senseid), f'{sk}--legacy senseid--> {legacy_senseid}')



"""
WordNet XML ID well-formedness selected cases tests
Author: John McCrae <john@mccr.ae> for original code
Author: Bernard Bou <1313ou@gmail.com> for rewrite and revamp
"""
#  Copyright (c) 2024.
#  Creative Commons 4 for original code
#  GPL3 for rewrite

import unittest

from oewn_xml.wordnet_xml import (
    escape_lemma, unescape_lemma, from_xml_entry_id, to_xml_entry_id,
    escape_sensekey, unescape_sensekey, to_xml_sense_id, from_xml_sense_id,
    legacy_factory, DashNameFactory, is_valid_xml_id, is_valid_xml_oewn_id)
from tests.utils import is_valid_xsd_id, make_dummy_sk


def try_is_valid_xsd_id(_id) -> None:
    try:
        is_valid_xsd_id(_id)
        print(f'{_id} is a valid xsd:id')
    except ValueError as ve:
        print(f'{_id} is NOT a valid xsd:id')
        raise ve


class XMLIDTestCase(unittest.TestCase):

    def expect_valid_xsd_id(self, _id) -> None:
        try_is_valid_xsd_id(_id)
        self.assertTrue(True)

    def expect_not_valid_xsd_id(self, _id) -> None:
        with self.assertRaises(ValueError, msg=f'{_id} is a valid xsd:id'):
            try_is_valid_xsd_id(_id)

    def test_rogue(self) -> None:
        print('\nROGUE')
        lemma = 'Capital: Critique of Political Economy'
        sk = 'capital:_critique_of_political_economy%1:10:01::'

        esc_lemma = escape_lemma(lemma)
        xml_entry_id = to_xml_entry_id(lemma, 'n')
        esc_sk = escape_sensekey(sk)
        xml_sense_id = to_xml_sense_id(sk)

        lemma_el = unescape_lemma(esc_lemma)
        lemma_eid, _, _ = from_xml_entry_id(xml_entry_id)
        sk_esk = unescape_sensekey(esc_sk)
        sk_sid = from_xml_sense_id(xml_sense_id)

        print(f'{lemma} --escape_lemma()--> {esc_lemma} --unescape_lemma()--> {lemma_el}')
        print(f'{sk} --escape_sensekey()--> {esc_sk} --unescape_sensekey()--> {sk_esk}')

        print(f'{lemma} --make_entry_id()--> {xml_entry_id} --unmake_entry_id()--> {lemma_eid}')
        print(f'{sk} --make_sense_id()--> {xml_sense_id} --unmake_sense_id()--> {sk_sid}')

        self.assertTrue(is_valid_xml_id(esc_lemma))
        self.assertTrue(is_valid_xml_id(esc_sk))
        self.assertTrue(is_valid_xml_id(xml_entry_id))
        self.assertTrue(is_valid_xml_id(xml_sense_id))

        self.assertTrue(is_valid_xml_oewn_id(xml_entry_id))
        self.assertTrue(is_valid_xml_oewn_id(xml_sense_id))

        self.assertEqual(lemma_el, lemma)
        self.assertEqual(lemma_eid, lemma)
        self.assertEqual(sk_esk, sk)
        self.assertEqual(sk_sid, sk)

    def test_rogue_xsd_id(self) -> None:
        print('\nXSD IDS - ROGUE')
        test_factory = DashNameFactory('__', '.')
        for lemma in ('Capital: Critique of Political Economy',):
            sk = make_dummy_sk(lemma)
            e_id = to_xml_entry_id(lemma, 'n', name_factory=test_factory)
            legacy_e_id = to_xml_entry_id(lemma, 'n', name_factory=legacy_factory)
            sk_id = to_xml_sense_id(sk, name_factory=test_factory)
            legacy_sk_id = to_xml_sense_id(sk, name_factory=legacy_factory)
            print(f'{lemma} -entryid--------> {e_id}')
            print(f'{lemma} -legacy entryid-> {legacy_e_id}')
            print(f'{sk} -senseid--------> {sk_id}')
            print(f'{sk} -legacy senseid-> {legacy_sk_id}')
            print()

            self.assertTrue(is_valid_xml_id(e_id))
            print(f'{e_id} is valid XML ID')
            self.assertTrue(is_valid_xml_id(e_id))
            print(f'{legacy_e_id} is valid XML ID')
            self.assertTrue(is_valid_xml_id(sk_id))
            print(f'{sk_id} is valid XML ID')
            self.assertTrue(is_valid_xml_id(legacy_sk_id))
            print(f'{legacy_sk_id} is valid XML ID')

            self.expect_valid_xsd_id(e_id)
            self.expect_valid_xsd_id(legacy_e_id)
            self.expect_valid_xsd_id(sk_id)
            self.expect_valid_xsd_id(legacy_sk_id)

    def test_xsd_id(self) -> None:
        print('\nXSD IDS')
        dummy_colon_id = 'dummy:dummy'
        with self.assertRaises(ValueError, msg=f'{dummy_colon_id} is a valid xsd:id'):
            try_is_valid_xsd_id(dummy_colon_id)
        test_factory = DashNameFactory('__', '.')
        for lemma in ('bass',):
            sk = make_dummy_sk(lemma)
            e_id = to_xml_entry_id(lemma, 'n')
            sk_id = to_xml_sense_id(sk, name_factory=test_factory)
            legacy_sk_id = to_xml_sense_id(sk, name_factory=legacy_factory)
            print(f'{lemma} -entryid-> {e_id}')
            print(f'{sk} -senseid--------> {sk_id}')
            print(f'{sk} -legacy senseid-> {legacy_sk_id}')
            print()

            self.assertTrue(is_valid_xml_id(e_id))
            print(f'{e_id} is valid XML ID')
            self.assertTrue(is_valid_xml_id(sk_id))
            print(f'{sk_id} is valid XML ID')
            self.assertTrue(is_valid_xml_id(legacy_sk_id))
            print(f'legacy {legacy_sk_id} is valid XML ID')

            self.expect_valid_xsd_id(e_id)
            self.expect_valid_xsd_id(sk_id)
            self.expect_valid_xsd_id(legacy_sk_id)

        print()

        for lemma in ('1:1', 'Capital: Critique of Political Economy'):
            sk = make_dummy_sk(lemma)
            e_id = to_xml_entry_id(lemma, 'n')
            sk_id = to_xml_sense_id(sk, name_factory=test_factory)
            legacy_sk_id = to_xml_sense_id(sk, name_factory=legacy_factory)
            print(f'{lemma} -entryid-> {e_id}')
            print(f'{sk} -senseid--------> {sk_id}')
            print(f'{sk} -legacy senseid-> {legacy_sk_id}')
            print()

            self.assertTrue(is_valid_xml_id(e_id))
            print(f'{e_id} is valid XML ID')
            self.assertTrue(is_valid_xml_id(sk_id))
            print(f'{sk_id} is valid XML ID')
            self.assertTrue(is_valid_xml_id(legacy_sk_id))
            print(f'{legacy_sk_id} is valid XML ID')

            self.expect_valid_xsd_id(e_id)
            self.expect_valid_xsd_id(sk_id)
            self.expect_valid_xsd_id(legacy_sk_id)


if __name__ == '__main__':
    unittest.main()



"""
WordNet utilities for tests
Author: John McCrae <john@mccr.ae> for original code
Author: Bernard Bou <1313ou@gmail.com> for rewrite and revamp
"""
#  Copyright (c) 2024.
#  Creative Commons 4 for original code
#  GPL3 for rewrite

from typing import Any, List, Dict, Tuple

from oewn_core.wordnet import Synset, Sense
from oewn_xml.wordnet_xml import dash_factory, legacy_factory, to_xml_sense_id, from_xml_sense_id, is_valid_xml_id, split_at_last


def collect_entries_for_escapes(entries, escape_map) -> Dict[Any, List]:
    r = {}
    for k in escape_map:
        if k == ' ':
            continue
        r[k] = []
        for e in entries:
            if k in e.lemma:
                r[k].append(e)
    return r


def print_as_dictionary(r, limit=5) -> None:
    print('escapable = {')
    for k, v in r.items():
        print(f'\t"{k}": (')
        i = 0
        for e in v:
            if i > limit:
                print(f'\t\t# ... {len(r[k])} total')
                break
            print(f'\t\t{e.key},')
            i += 1
        print('\t),')
    print('}')


def dump(s: Synset | Sense):
    print(s)
    for r in sorted(s.relations, key=lambda r2: (r2.relation_type, r2.target)):
        print(f'\t{r}')


def make_dummy_sk(lemma) -> str:
    return make_sk(lemma, '1:00:00::')


def make_sk(lemma, lex_sense) -> str:
    return f'{lemma.lower().replace(' ', '_')}%{lex_sense}'


def is_valid_xsd_id(s) -> bool:
    if not is_valid_xml_id(s):
        raise ValueError(f'Invalid xsd:id : {s} is not a valid XML ID')
    if ':' in s:
        raise ValueError(f"Invalid xsd:id : {s} contains colon ':'")
    return True


def is_parsable_sensekey(sk) -> bool:
    try:
        l, s = split_at_last(sk, '%')
        f = s.split(':')
        if len(f) == 5:
            return True
        raise ValueError(f'COLON: {sk.id}')
    except ValueError as ve:
        raise ValueError(f'PERCENT: {sk.id} {ve}')


def is_parsable_xml_sensekey(sk) -> bool:
    try:
        l, s = split_at_last(sk, dash_factory.xml_percent_sep)
        f = s.split(dash_factory.xml_colon_sep)
        if len(f) == 5:
            return True
        raise ValueError(f'COLON: {sk.id}')
    except ValueError as ve:
        raise ValueError(f'PERCENT: {sk.id} {ve}')


def generate_senseid_from_sensekey(sk: str) -> Tuple[str, str, str]:
    senseid = to_xml_sense_id(sk)
    sk2 = from_xml_sense_id(senseid)
    if sk != sk2:
        raise ValueError(f'unescaped != original: {sk} != {sk2}')
    return sk, senseid, sk2


def generate_legacy_senseid_from_sensekey(sk) -> Tuple[str, str, str]:
    senseid = to_xml_sense_id(sk, legacy_factory)
    sk2 = from_xml_sense_id(senseid, legacy_factory)
    if sk != sk2:
        raise ValueError(f'unescaped != original: {sk} != {senseid}')
    return sk, senseid, sk2



"""
WordNet to-YAML utilities

Author: John McCrae <john@mccr.ae> for original code
Author: Bernard Bou <1313ou@gmail.com> for rewrite and revamp
"""

#  Copyright (c) 2024.
#  Creative Commons 4 for original code
#  GPL3 for rewrite

import codecs
from typing import Dict, List, Any

import yaml

from oewn_core.wordnet import WordnetModel, Sense, Synset, Example, ignored_symmetric_sense_relations, ignored_symmetric_synset_relations

az = 'abcdefghijklmnopqrstuvwxyz'

check_resolved = False
""" Whether resolved_* members' resolution is checked, a no-op because these are not saved """


def entry_to_yaml(entry, sense_resolver=None) -> Dict[str, Any]:
    """
    Build dictionary for lexical entry YAML
    :param entry: lexical entry
    :param sense_resolver: if not None, sense resolution will we attempted
    :return: dictionary
    """
    y = {}
    if entry.forms:
        y['form'] = entry.forms
    if entry.pronunciations:
        y['pronunciation'] = []
        for p in entry.pronunciations:
            if p.variety:
                y['pronunciation'].append({'value': p.value, 'variety': p.variety})
            else:
                y['pronunciation'].append({'value': p.value})
    y['sense'] = [sense_to_yaml(s, sense_resolver) for s in entry.senses]
    return y


def sense_to_yaml(sense, sense_resolver=None) -> Dict[str, Any]:
    """
    Build dictionary for sense YAML
    :param sense: sense
    :param sense_resolver: if not None, sense resolution will be attempted and checked
    :return: dictionary
    """
    y = {'id': sense.id, 'synset': sense.synsetid, }
    if sense.adjposition:
        y['adjposition'] = sense.adjposition
    if sense.verbframeids:
        y['subcat'] = sense.verbframeids
    if sense.examples:
        y['sent'] = sense.examples
    yr: Dict[str, List[str]] = sense_relations_to_yaml(sense, sense_resolver)
    if yr:
        y.update(yr)
    return y


def synset_to_yaml(synset, synset_resolver=None, member_resolver=None) -> Dict[str, Any]:
    """
    Build dictionary for synset YAML
    :param synset: synset
    :param synset_resolver: if not None, synset resolution will be attempted and checked
    :param member_resolver: if not None, member resolution will be attempted and checked
    :return: dictionary
    """
    if member_resolver and not all((m, synset.id) in member_resolver for m in synset.members):
        raise ValueError(f'Unresolved member in {synset.members}')

    y = {
        'members': synset.members,
        'partOfSpeech': synset.pos,
        'definition': synset.definitions
    }
    if synset.examples:
        y['example'] = [example_to_yaml(x) for x in synset.examples]
    if synset.usages:
        y['usage'] = synset.usages
    if synset.wikidata:
        y['wikidata'] = synset.wikidata
    if synset.source:
        y['source'] = synset.source
    if synset.ili and synset.ili != 'in':
        y['ili'] = synset.ili
    yr: Dict[str, List[str]] = synset_relations_to_yaml(synset, synset_resolver)
    if yr:
        y.update(yr)
    return y


def sense_relations_to_yaml(sense: Sense, sense_resolver=None) -> Dict[str, List[str]]:
    y: Dict[str, List[str]] = {}
    for r in sense.relations:
        t = Sense.Relation.OtherType(r.relation_type) if r.other_type else Sense.Relation.Type(r.relation_type)
        if t not in ignored_symmetric_sense_relations:
            if sense_resolver and (r.target not in sense_resolver):
                raise ValueError(f'Unresolved sense relation target {r.target} of type {t.value} in {sense.id}')
            if check_resolved and sense_resolver and (
                    r.target not in sense_resolver or sense_resolver[r.target] != r.resolved_target):
                raise ValueError(f'Invalid sense relation resolved target {r.target} of type {t.value} in {sense.id}')
            k: str = str(t.value)
            if k not in y:
                y[k] = []
            y[k].append(r.target)
    return y


def synset_relations_to_yaml(synset: Synset, synset_resolver=None) -> Dict[str, List[str]]:
    y: Dict[str, List[str]] = {}
    for r in synset.relations:
        t = Synset.Relation.Type(r.relation_type)
        if t not in ignored_symmetric_synset_relations:
            if synset_resolver and r.target not in synset_resolver:
                raise ValueError(f'Unresolved synset relation target {r.target} of type {t.value} in {synset.id}')
            if check_resolved and synset_resolver and (
                    r.target not in synset_resolver or synset_resolver[r.target] != r.resolved_target):
                raise ValueError(f'Invalid synset relation resolved target {r.target} of type {t.value} in {synset.id}')
            k: str = str(t.value)
            if t.value not in y:
                y[k] = []
            y[k].append(r.target)
    return y


def example_to_yaml(example: str | Example) -> Dict[str, str] | str:
    """
    Build dictionary for sourced example or text for unsourced example YAML
    :param example: example
    :return: dictionary or string
    """
    if isinstance(example, Example):
        return {'text': example.text, 'source': example.source} if example.source else example.text
    return example


def save_entries(wn: WordnetModel, home: str) -> None:
    """
    Persist entries to YAML (entries-(0|a|...|z).yaml
    :param wn: model
    :param home: home dir for persist files
    :return: None
    """
    sense_resolver = wn.sense_resolver if wn.sense_resolver else None
    entry_yaml = {c: {} for c in az}
    entry_yaml['0'] = {}
    for entry in wn.entries:
        # build
        y = entry_to_yaml(entry, sense_resolver=sense_resolver)

        # locate
        first = entry.lemma.lower()[:1]
        if first not in az:
            first = '0'

        # super dict
        if entry.lemma not in entry_yaml[first]:
            entry_yaml[first][entry.lemma] = {}

        # uniqueness
        key = f'{entry.pos}-{entry.discriminant}' if entry.discriminant else entry.pos
        if key in entry_yaml[first][entry.lemma]:
            raise ValueError(f'Duplicate entry: {entry.lemma}-{key}')

        entry_yaml[first][entry.lemma][key] = y

    # save
    for c in az:
        with codecs.open(f'{home}/entries-%s.yaml' % c, 'w', 'utf-8') as out:
            yaml.dump(entry_yaml[c], out, allow_unicode=True)
    with codecs.open(f'{home}/entries-0.yaml', 'w', 'utf-8') as out:
        yaml.dump(entry_yaml['0'], out, allow_unicode=True)


def save_synsets(wn: WordnetModel, home: str) -> None:
    """
    Persist synsets to YAML (noun|verb|adj|adv)*.yaml
    :param wn: model
    :param home: home dir for persist files
    """
    synset_yaml = {}
    for synset in wn.synsets:
        # build
        y = synset_to_yaml(synset, synset_resolver=wn.synset_resolver, member_resolver=wn.member_resolver)

        # super dict
        if synset.lex_name not in synset_yaml:
            synset_yaml[synset.lex_name] = {}

        # uniqueness
        if synset.id in synset_yaml[synset.lex_name]:
            raise ValueError(f'Duplicate synset: {synset.id} in {synset.lex_name}')
        synset_yaml[synset.lex_name][synset.id] = y

    # save
    for key, synsets in synset_yaml.items():
        with codecs.open(f'{home}/%s.yaml' % key, 'w', 'utf-8') as out:
            yaml.dump(synsets, out, allow_unicode=True)


def save_verbframes(wn: WordnetModel, home: str) -> None:
    """
    Persist verb frames to YAML frame.yaml
    :param wn: model
    :param home: home dir for persist file
     """
    frame_yaml = {b.id: b.verbframe for b in wn.verbframes}
    with open(f'{home}/frames.yaml', 'w', encoding='utf-8') as out:
        yaml.dump(frame_yaml, out, allow_unicode=True)


def save(wn: WordnetModel, home: str) -> None:
    """
    Persist model to YAML *.yaml
    :param wn: model
    :param home: home dir for persist files
    """
    print(f'saving to YAML {home}')
    save_entries(wn, home)
    save_synsets(wn, home)
    save_verbframes(wn, home)
    print(f'saved to YAML {home}')



"""
WordNet sensekey tests
Author: John McCrae <john@mccr.ae> for original code
Author: Bernard Bou <1313ou@gmail.com> for rewrite and revamp
"""
#  Copyright (c) 2024.
#  Creative Commons 4 for original code
#  GPL3 for rewrite

import sys
import unittest

from oewn_xml.wordnet_xml import is_valid_xml_id, to_xml_sense_id, from_xml_sense_id, key_prefix_len
from tests.model import sorted_senses
from tests.utils import is_parsable_sensekey, is_parsable_xml_sensekey, is_valid_xsd_id


class SensekeysTestCase(unittest.TestCase):

    def test_sensekeys(self) -> None:
        for s in sorted_senses:
            try:
                self.assertTrue(is_parsable_sensekey(s.id))
                sid = to_xml_sense_id(s.id)
                self.assertTrue(is_valid_xml_id(sid), f'{sid}')
                self.assertTrue(is_valid_xsd_id(sid), f'{sid}')
                self.assertTrue(is_parsable_xml_sensekey(sid[key_prefix_len:]), f'{sid}')
                sk = from_xml_sense_id(sid)
                self.assertEqual(s.id, sk)
                print(f'{s.id}   {sid}   {sk}')
            except ValueError as e:
                print(e, file=sys.stderr)


if __name__ == '__main__':
    unittest.main()



"""
WordNet model extend tests
Author: John McCrae <john@mccr.ae> for original code
Author: Bernard Bou <1313ou@gmail.com> for rewrite and revamp
"""
#  Copyright (c) 2024.
#  Creative Commons 4 for original code
#  GPL3 for rewrite

import unittest

from tests.model import wn
from tests.utils import dump


class ExtendTestCase(unittest.TestCase):

    def test_extend(self) -> None:
        s13 = wn.synset_resolver['02372362-v']
        s12 = wn.synset_resolver['00772482-v']
        s11 = wn.synset_resolver['02512195-v']
        s23 = wn.synset_resolver['05544491-n']
        s22 = wn.synset_resolver['05543502-n']
        s21 = wn.synset_resolver['05543117-n']
        print("\nBEFORE EXTEND")
        for s in (s11, s12, s13, s21, s22, s23):
            dump(s)

        wn.extend()

        print("\nAFTER EXTEND")
        for s in (s11, s12, s13, s21, s22, s23):
            dump(s)
        self.assertTrue(True)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python3

"""
WordNet serialize

Author: John McCrae <john@mccr.ae> for original code
Author: Bernard Bou <1313ou@gmail.com> for rewrite and revamp
"""

#  Copyright (c) 2024.
#  Creative Commons 4 for original code
#  GPL3 for rewrite

import argparse
import pickle
import sys
import time

from oewn_core.wordnet import WordnetModel
from oewn_core.wordnet_fromyaml import load


def save_pickle(wn: WordnetModel, path: str, file: str = 'wn.pickle', verbose: bool = False) -> None:
    """
    Save model to pickle file in path
    Cross-references don't have to be staled.
    """
    if verbose:
        print(f'saving to pickle in {path}/{file}')
    with open(f'{path}/{file}', 'wb') as out:
        pickle.dump(wn, out)
    if verbose:
        print(f'saved to pickle in {path}/{file}')


def main() -> WordnetModel:
    """
    WordNet load-save
    Will have a normalizing effect, after which it's not modified
    """
    arg_parser = argparse.ArgumentParser(description="load from yaml and save")
    arg_parser.add_argument('in_dir', type=str, help='from-dir')
    arg_parser.add_argument('out_dir', type=str, help='to-dir for pickle')
    arg_parser.add_argument('pickled', type=str, nargs='?', default='oewn.pickle', help='to-pickle')
    args = arg_parser.parse_args()

    wn = load(args.in_dir)
    save_pickle(wn, args.out_dir, args.pickled)
    return wn


def test(out_dir, verbose: bool = False) -> None:
    from oewn_core.deserialize import load_pickle
    if verbose:
        print(f'loading from pickle in {out_dir}')
    wn = load_pickle(out_dir)
    if verbose:
        print(f'loading from pickle in {out_dir}')
        print(wn)
        print(wn.info())
        print(wn.info_relations())


if __name__ == '__main__':
    start_time = time.time()
    main()
    end_time = time.time()
    duration = end_time - start_time
    print(f"Saving took {duration:.6f} seconds", file=sys.stderr)



#!/usr/bin/python3

"""
Export SyntagNet collocation in OEWN YAML format
lex:
  pos-discriminant:
    sense:
      - collocation:
        - sensekey2a
        - sensekey2b
        ...
        id: sensekey1

abandon:
  v:
    sense:
    - collocation:
      - 'baby%1:18:00::'
      - 'car%1:06:00::'
      - 'favor%1:07:00::'
      - 'project%1:09:00::'
      - 'vehicle%1:06:00::'
      id: 'abandon%2:40:00::'

Author: Bernard Bou <1313ou@gmail.com> for rewrite and revamp
"""

#  Copyright (c) 2024.
#  Creative Commons 4 for original code
#  GPL3 for rewrite

import argparse
import codecs
import sys
import time
from typing import Dict, List, Any, Optional

import yaml

from oewn_core.wordnet import WordnetModel, Sense
from oewn_syntagnet.inject_syntagnet import load_and_inject


def sense_relations_to_yaml(sense: Sense) -> List[str]:
    """
    Build YAML dictionary for sense relation
    :param sense: sense
    :return: YAML dictionary or None if sense has no collocation
    """

    y: List[str] = []
    for r in sense.relations:
        if not r.other_type and Sense.Relation.Type(r.relation_type) == Sense.Relation.Type.COLLOCATION:
            y.append(r.target)
    return y


def sense_to_yaml(sense) -> Optional[Dict[str, Any]]:
    """
    Build YAML dictionary for sense
    :param sense: sense
    :return: YAML dictionary or None if sense has no collocation
    """

    yr: List[str] = sense_relations_to_yaml(sense)
    if yr:
        y: Dict[str, Any] = {'id': sense.id, 'collocation': yr}
        return y
    return None


def entry_to_yaml(entry) -> Optional[Dict[str, Any]]:
    """
    Build YAML dictionary for lexical entry
    :param entry: lexical entry
    :return: YAML dictionary
    """
    yss = [ys for s in entry.senses if (ys := sense_to_yaml(s)) is not None]
    if yss:
        return {'sense': yss}
    return None


def save_entries(wn: WordnetModel, home: str) -> None:
    """
    Persist entries to YAML (entries-(0|a|...|z).yaml
    :param wn: model
    :param home: home dir for persist files
    :return: None
    """
    y = {}
    for entry in wn.entries:
        # build
        ye = entry_to_yaml(entry)
        if ye:
            # first-tier key
            if entry.lemma not in y:
                y[entry.lemma] = {}
            # second-tier key
            key = f'{entry.pos}-{entry.discriminant}' if entry.discriminant else entry.pos
            if key not in y[entry.lemma]:
                y[entry.lemma][key] = {}

            y[entry.lemma][key] = ye

    # save
    with codecs.open(f'{home}/oewn-syntagnet.yaml', 'w', 'utf-8') as out:
        yaml.dump(y, out, allow_unicode=True)


def save(wn: WordnetModel, home: str) -> None:
    """
    Persist model to YAML syntagnet.yaml
    :param wn: model
    :param home: home dir for persist files
    """
    print(f'saving to YAML {home}')
    save_entries(wn, home)
    print(f'saved to YAML {home}')


def main() -> None:
    """
    Export SyntagNet as YAML
    """
    arg_parser = argparse.ArgumentParser(description="load wn and syntagnet from yaml, merge and save")
    arg_parser.add_argument('--pickle', action='store_true', default=False, help='use pickle')
    arg_parser.add_argument('in_dir', type=str, help='from-dir')
    arg_parser.add_argument('out_dir', type=str, help='to-dir')
    arg_parser.add_argument('syntagnet', type=str, help='collocations')
    arg_parser.add_argument('pickled', type=str, nargs='?', default=None, help='from-pickle')
    args = arg_parser.parse_args()

    wn: WordnetModel = load_and_inject(args.in_dir, args.syntagnet, args.pickled if args.pickle else None)
    save(wn, args.out_dir)


if __name__ == '__main__':
    start_time = time.time()
    main()
    end_time = time.time()
    duration = end_time - start_time
    print(f"Exporting took {duration:.6f} seconds", file=sys.stderr)



"""
WordNet model transitive validation tests
When a--hypernym-->b--hypernym-->c, there shouldn't be a--hypernym-->c
Author: John McCrae <john@mccr.ae> for original code
Author: Bernard Bou <1313ou@gmail.com> for rewrite and revamp
"""
#  Copyright (c) 2024.
#  Creative Commons 4 for original code
#  GPL3 for rewrite

import unittest

from oewn_core.wordnet import Synset
from oewn_validate.validate import check_transitive_synset, ValidationError
from tests.model import wn
from tests.utils import dump


class TransitiveTestCase(unittest.TestCase):

    def test_transitive(self) -> None:
        s1: Synset = wn.synset_resolver['05543117-n']
        # s2: Synset = wn.synset_resolver['05543502-n']
        s3: Synset = wn.synset_resolver['05544491-n']

        wn.extend()

        check_transitive_synset(wn, s3)

        print("\nBEFORE EXTEND")
        dump(s3)
        r = Synset.Relation(s1.id, Synset.Relation.Type.HYPERNYM.value)
        s3.relations.append(r)
        print("\nAFTER EXTEND")
        dump(s3)

        with self.assertRaises(ValidationError):
            check_transitive_synset(wn, s3)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python3

"""
Inject SyntagNet (YAML) into the model

Author: Bernard Bou <1313ou@gmail.com> for rewrite and revamp
"""

#  Copyright (c) 2024.
#  Creative Commons 4 for original code
#  GPL3 for rewrite

import argparse
import sys
import time
from typing import Dict, Any, Set, Tuple, Optional

import yaml

from oewn_core.wordnet import WordnetModel, Sense


def make_sensekeys(wn: WordnetModel) -> Set[str]:
    """
    :param wn: model
    :return: set of sensekeys in the model
    """
    sensekeys: Set = set()
    for s in wn.senses:
        sensekeys.add(s.id)
    return sensekeys


def inject_syntagnet_to_model(wn: WordnetModel, syntagnet: str, two_ways: bool = True) -> Tuple[int, int]:
    """
    WordNet inject SyntagNet as relations
    :param wn: model
    :param syntagnet: path to SyntagNet YAML data
    :param two_ways: whether to add inverse relations also
    :return: count of additions and failures
    """
    count = 0
    with open(syntagnet, encoding='utf-8') as inp:
        fails = 0
        y: Dict[str, Any] = yaml.load(inp, Loader=yaml.CLoader)
        for sk1, collocations_y in y.items():
            # print(f'{sk} {collocations_y}')
            if sk1 in wn.sense_resolver:
                sense1 = wn.sense_resolver[sk1]
                # print(f'{sense1.id}')
                for sk2 in collocations_y["collocation"]:
                    if sk2 in wn.sense_resolver:
                        sense2 = wn.sense_resolver[sk2]
                        # print(f'\t{sense1.id}')

                        # type
                        t = Sense.Relation.Type.COLLOCATION

                        # add to sense 1
                        if sense1.relations is None:
                            sense1.relations = []
                        # print(f"add {sk1}-{t}->{sk2}")
                        sense1.relations.append(Sense.Relation(sk2, t))
                        count += 1

                        if two_ways:
                            # add to sense 2
                            if sense2.relations is None:
                                sense2.relations = []
                            # print(f"add {sk2}-{t}->{sk1}")
                            sense2.relations.append(Sense.Relation(sk1, t))
                            count += 1
                else:
                    print(f'{sk2} target not resolvable in collocation {sk1}-{sk2}', file=sys.stderr)
                    fails += 1

            else:
                print(f'{sk1} source not resolvable in collocation for {sk1}', file=sys.stderr)
                fails += 1
            sense1.relations = list(sorted(set(sense1.relations), key=lambda r: (r.relation_type, r.target)))
            if two_ways:
                sense2.relations = list(sorted(set(sense2.relations), key=lambda r: (r.relation_type, r.target)))

    return count, fails


def load_and_inject(in_dir: str, syntagnet: str, pickled: Optional[str] = None) -> WordnetModel:
    """
    Load Wordnet and SyntagNet from YAML, inject sn into wn
    :param in_dir: home dir for YAML or pickled model file(s)
    :param syntagnet: path to SyntagNet YAML data
    :param pickled: whether to use pickled model
    :return: Syntagnet-augmented model
    """

    def get_model() -> WordnetModel:
        if pickled:
            from oewn_core.deserialize import load
            return load(in_dir, file=pickled)
        else:
            from oewn_core.wordnet_fromyaml import load
            return load(in_dir, resolve=True)

    wn = get_model()
    count, fails = inject_syntagnet_to_model(wn, syntagnet)
    print(f'{count} collocations {fails} fails', file=sys.stderr)
    return wn


def main() -> None:
    """
    WordNet + SyntagNet load-save
    :command-line: oewn_home output_home syntagnet_file
    """
    from oewn_core.wordnet_toyaml import save
    arg_parser = argparse.ArgumentParser(description="load wn and SyntagNet from yaml, inject and save")
    arg_parser.add_argument('--pickle', action='store_true', default=False, help='use pickle')
    arg_parser.add_argument('in_dir', type=str, help='from-dir')
    arg_parser.add_argument('out_dir', type=str, help='to-dir')
    arg_parser.add_argument('syntagnet', type=str, help='collocations')
    arg_parser.add_argument('pickled', type=str, nargs='?', default=None, help='from-pickle')
    args = arg_parser.parse_args()
    wn = load_and_inject(args.in_dir, args.syntagnet, args.pickled if args.pickle else None)
    save(wn, args.out_dir)


if __name__ == '__main__':
    start_time = time.time()
    main()
    end_time = time.time()
    duration = end_time - start_time
    print(f"Injecting took {duration:.6f} seconds", file=sys.stderr)



#!/usr/bin/python3

"""
Augment YAML with SyntagNet collocations
Merges exported OEWN-SyntagNet into OEWN YAML files

Author: Bernard Bou <1313ou@gmail.com> for rewrite and revamp
"""

#  Copyright (c) 2024.
#  Creative Commons 4 for original code
#  GPL3 for rewrite

import argparse
import os
import sys
from glob import glob
from typing import Dict, Any, List

import yaml


def augment_yaml_files(out_dir, incoming: Dict[str, Any], files: List[str]):
    for file in files:
        name = os.path.basename(file)
        print(f'{name}', file=sys.stderr)
        with open(file) as f:
            base = yaml.safe_load(f)
            _deep_augment(base, incoming, 0)

            with open(f'{out_dir}/{name}', 'w') as out:
                yaml.dump(base, out, allow_unicode=True)


def _deep_augment(base: Dict, incoming: Dict, level: int) -> Dict:
    for key, value in base.items():
        if key in incoming:
            value2 = incoming[key]
            if isinstance(value, dict) and isinstance(value2, dict):
                _deep_augment(value, value2, level + 1)
            elif isinstance(value, list) and isinstance(value2, list):
                dd = {d['id']: d for d in value2}
                for d in value:
                    i = d['id']
                    if i in dd:
                        d2 = dd[i]
                        d.update(d2)
                # print(f'{'  ' * level}{key} - {value}')
    return base


def main() -> None:
    arg_parser = argparse.ArgumentParser(description="load syntagnet from yaml, merge it into yaml files")
    arg_parser.add_argument('in_dir', type=str, help='from-dir for yaml/pickle')
    arg_parser.add_argument('out_dir', type=str, help='out-dir for result')
    arg_parser.add_argument('syntagnet', type=str, nargs='?', default='syntagnet.yaml', help='syntagnet data')
    args = arg_parser.parse_args()

    with open(args.syntagnet) as f:
        print(f'{os.path.basename(args.syntagnet)}', file=sys.stderr)
        incoming = yaml.safe_load(f)
        if isinstance(incoming, dict):
            augmentable_files = sorted(glob(f'{args.in_dir}/entries-*.yaml'))
            augment_yaml_files(f'{args.out_dir}', incoming, augmentable_files)
        else:
            print("Cannot augment with non dictionaries {}", file=sys.stderr)


if __name__ == '__main__':
    main()



"""
WordNet model validation
This is validation of model's semantics
This validation does not validate XML form (use an XML validator)

Author: John McCrae <john@mccr.ae> for original code
Author: Bernard Bou <1313ou@gmail.com> for rewrite and revamp
"""
#  Copyright (c) 2024.
#  Creative Commons 4 for original code
#  GPL3 for rewrite

import argparse
import re
import sys
from collections import Counter
from typing import Dict, Pattern, Tuple

from oewn_core.wordnet import Entry, Synset, Sense, PartOfSpeech, WordnetModel


class ValidationError(Exception):
    def __init__(self, message) -> None:
        self.message = message
        super().__init__(self.message)


break_on_error = True


def warn(message: str) -> None:
    if break_on_error:
        raise ValidationError(message)
    else:
        print(message, file=sys.stderr)


# I D   V A L I D A T I O N

valid_id: Pattern = re.compile(fr"^.+$")

# valid_lemma = f'[^ %:]+'
valid_lemma: str = f'[^ %]+'

# valid_head = f'[^ %:]*'
valid_head: str = f'[^ %]*'

valid_sense_id: Pattern = re.compile(fr'^{valid_lemma}%([0-9]):[0-9]{{2}}:[0-9]{{2}}:{valid_head}:[0-9]{{0,2}}$')

valid_synset_id: Pattern = re.compile('^[0-9]{8}-([nvars])$')


def check_valid_id(any_id: str) -> None:
    if not bool(valid_id.match(any_id)):
        warn(f'{any_id} is not well-formed id')


def check_valid_synset_id(synset_id: str) -> None:
    check_valid_id(synset_id)
    if not bool(valid_synset_id.match(synset_id)):
        warn(f'{synset_id} is not well-formed synset id')


def check_valid_sense_id(sense_id: str) -> None:
    check_valid_id(sense_id)
    if not bool(valid_sense_id.match(sense_id)):
        warn(f'{sense_id} is not well-formed sense id')


def check_valid_sense_id_for_target(sense_id: str, target_id: str) -> None:
    m = valid_sense_id.match(sense_id)
    assert m is not None
    g = m.group(1)
    pos = ss_types_reverse[int(g)]
    m2 = valid_synset_id.match(target_id)
    assert m2 is not None
    g2 = m2.group(1)
    pos2 = PartOfSpeech(g2)
    if pos != pos2:
        warn(f'{target_id} target of {sense_id} is not well-formed')


# U T I L I T I E S   F O R   S E N S E K E Y S

lex_file_nums: Dict[str, int] = {
    "adj.all": 0,
    "adj.pert": 1,
    "adv.all": 2,
    "noun.Tops": 3,
    "noun.act": 4,
    "noun.animal": 5,
    "noun.artifact": 6,
    "noun.attribute": 7,
    "noun.body": 8,
    "noun.cognition": 9,
    "noun.communication": 10,
    "noun.event": 11,
    "noun.feeling": 12,
    "noun.food": 13,
    "noun.group": 14,
    "noun.location": 15,
    "noun.motive": 16,
    "noun.object": 17,
    "noun.person": 18,
    "noun.phenomenon": 19,
    "noun.plant": 20,
    "noun.possession": 21,
    "noun.process": 22,
    "noun.quantity": 23,
    "noun.relation": 24,
    "noun.shape": 25,
    "noun.state": 26,
    "noun.substance": 27,
    "noun.time": 28,
    "verb.body": 29,
    "verb.change": 30,
    "verb.cognition": 31,
    "verb.communication": 32,
    "verb.competition": 33,
    "verb.consumption": 34,
    "verb.contact": 35,
    "verb.creation": 36,
    "verb.emotion": 37,
    "verb.motion": 38,
    "verb.perception": 39,
    "verb.possession": 40,
    "verb.social": 41,
    "verb.stative": 42,
    "verb.weather": 43,
    "adj.ppl": 44,
    "contrib.colloq": 50,
    "contrib.plwn": 51}

ss_types: Dict[PartOfSpeech, int] = {
    PartOfSpeech.NOUN: 1,
    PartOfSpeech.VERB: 2,
    PartOfSpeech.ADJECTIVE: 3,
    PartOfSpeech.ADVERB: 4,
    PartOfSpeech.ADJECTIVE_SATELLITE: 5
}
ss_types_reverse: Dict[int, PartOfSpeech] = {ss_types[k]: k for k in ss_types}

sense_id_lex_id: Pattern[str] = re.compile(".*%\\d:\\d\\d:(\\d\\d):.*")


def gen_lex_id(entry: Entry, sense: Sense) -> int:
    max_id = 0
    unseen = 1
    seen = False
    for s2 in entry.senses:
        if s2.id:
            m = re.match(sense_id_lex_id, s2.id)
            assert m is not None
            max_id = max(max_id, int(m.group(1)))
        else:
            if not seen:
                if s2.id == sense.id:
                    seen = True
                else:
                    unseen += 1
    return max_id + unseen


def extract_lex_id(sense_key: str) -> int:
    m = re.match(sense_id_lex_id, sense_key)
    assert m is not None
    return int(m.group(1))


def get_head_word(wn: WordnetModel, sense: Sense) -> Tuple[str, str] | None:
    synset = wn.synset_resolver[sense.synsetid]
    similars = [r for r in synset.relations if
                Synset.Relation.Type(r.relation_type) == Synset.Relation.Type.SIMILAR and
                # ignore satellites in non-Princeton sets
                not r.target.startswith("9") and
                not r.target.startswith("8")]
    if len(similars) == 1:
        target_id = similars[0].target
        target = wn.synset_resolver[target_id]
        target_entry = wn.member_resolver[(target.members[0], target.id)]
        target_sense = next((s for s in target_entry.senses if s.synsetid == target_id), None)
        assert target_sense is not None
        head = target_sense.id[:target_sense.id.rindex('%')]
        m = re.match(sense_id_lex_id, target_sense.id)
        assert m is not None
        head_id = m.group(1)
        return head, head_id
    warn(f'Could not deduce target of satellite {sense.id}')


def make_sense_key(wn: WordnetModel, entry: Entry, sense: Sense) -> str:
    """Calculate the sense key for a sense of an entry"""
    lemma = entry.lemma.lower().replace(' ', '_')
    ss = wn.synset_resolver[sense.synsetid]
    pos = PartOfSpeech(ss.pos)
    ss_type = ss_types[pos]
    lex_filenum = lex_file_nums[ss.lex_name]
    if sense.id:
        lex_id = extract_lex_id(sense.id)
    else:
        lex_id = gen_lex_id(entry, sense)
    if pos == PartOfSpeech.ADJECTIVE_SATELLITE:
        h = get_head_word(wn, sense)
        assert h is not None
        head_word, head_id = h
    else:
        head_word = ""
        head_id = ""
    return f'{lemma}%{ss_type:1}:{lex_filenum:02}:{lex_id:02}:{head_word}:{head_id}'


def equal_pos(pos1: PartOfSpeech, pos2: PartOfSpeech) -> bool:
    return (pos1 == pos2
            or pos1 == PartOfSpeech.ADJECTIVE and pos2 == PartOfSpeech.ADJECTIVE_SATELLITE
            or pos2 == PartOfSpeech.ADJECTIVE and pos1 == PartOfSpeech.ADJECTIVE_SATELLITE)


# S T R U C T U R E  (R E L A T I O N S)

def check_symmetry_synset(wn: WordnetModel, synset: Synset) -> None:
    for r in synset.relations:
        t = Synset.Relation.Type(r.relation_type)
        if t in Synset.Relation.inverses:
            t2 = Synset.Relation.inverses[t]
            synset2 = wn.synset_resolver[r.target]
            if not any(r for r in synset2.relations if r.target == synset.id and Synset.Relation.Type(r.relation_type) == t2):
                warn(f'No symmetric relation for {synset.id} ={r.relation_type}=> {synset2.id}')


def check_symmetry_sense(wn: WordnetModel, sense: Sense) -> None:
    for r in sense.relations:
        if not r.other_type:
            t = Sense.Relation.Type(r.relation_type)
            if t in Sense.Relation.inverses:
                t2 = Sense.Relation.inverses[t]
                sense2 = wn.sense_resolver[r.target]
                if not any(r for r in sense2.relations if
                           r.target == sense.id and
                           not r.other_type and
                           Sense.Relation.Type(r.relation_type) == t2):
                    warn(f'No symmetric relation for {sense.id} ={r.relation_type}=> {sense2.id}')


def check_symmetry(wn: WordnetModel) -> None:
    for synset in wn.synsets:
        check_symmetry_synset(wn, synset)
    for sense in wn.senses:
        check_symmetry_sense(wn, sense)


def check_transitive_synset(wn: WordnetModel, synset: Synset) -> None:
    for r1 in synset.relations:
        t1 = Synset.Relation.Type(r1.relation_type)
        if t1 == Synset.Relation.Type.HYPERNYM:
            synset2 = wn.synset_resolver[r1.target]
            for r2 in synset2.relations:
                t2 = Synset.Relation.Type(r2.relation_type)
                if t2 == Synset.Relation.Type.HYPERNYM:
                    if any(r for r in synset.relations if
                           r.target == r2.target and Synset.Relation.Type(r.relation_type) == Synset.Relation.Type.HYPERNYM):
                        trans = next(r for r in synset.relations if
                                     r.target == r2.target and
                                     Synset.Relation.Type(r.relation_type) == Synset.Relation.Type.HYPERNYM)
                        warn(f'Transitive error for {synset.id} => {synset2.id} => {r2.target} with {synset.id} => {trans.target}')


def check_transitive(wn: WordnetModel) -> None:
    for synset in wn.synsets:
        check_transitive_synset(wn, synset)


def check_no_loops(wn: WordnetModel) -> None:
    hypernyms = {}
    for synset in wn.synsets:
        hypernyms[synset.id] = set()
        for r in synset.relations:
            if Synset.Relation.Type(r.relation_type) == Synset.Relation.Type.HYPERNYM:
                hypernyms[synset.id].add(r.target)
    changed = True
    while changed:
        changed = False
        for synset in wn.synsets:
            n_size = len(hypernyms[synset.id])
            for c in hypernyms[synset.id]:
                hypernyms[synset.id] = hypernyms[synset.id].union(hypernyms.get(c, []))
                if synset.id in hypernyms[synset.id]:
                    warn(f'Loop for {synset.id} <-> {c}')
            if len(hypernyms[synset.id]) != n_size:
                changed = True


def check_no_domain_loops(wn: WordnetModel) -> None:
    domains = {}
    for synset in wn.synsets:
        domains[synset.id] = set()
        for r in synset.relations:
            t = Synset.Relation.Type(r.relation_type)
            if t in (Synset.Relation.Type.DOMAIN_TOPIC, Synset.Relation.Type.DOMAIN_REGION, Synset.Relation.Type.EXEMPLIFIES):
                domains[synset.id].add(r.target)
    changed = True
    while changed:
        changed = False
        for synset in wn.synsets:
            n_size = len(domains[synset.id])
            for c in domains[synset.id]:
                domains[synset.id] = domains[synset.id].union(domains.get(c, []))
            if len(domains[synset.id]) != n_size:
                changed = True
            if synset.id in domains[synset.id]:
                warn(f'Domain loop for {synset.id}')


# E N T R I E S

def check_entry_keys(wn) -> None:
    entry_keys = set()
    for entry in wn.entries:
        k = entry.key
        if k in entry_keys:
            warn(f'Duplicate key: {k} for {entry}')
        entry_keys.add(k)


def check_entry_sense_duplicates(entry: Entry) -> None:
    for sense in entry.senses:
        for sense2 in entry.senses:
            if sense2.id != sense.id and sense2.synsetid == sense.synsetid:
                warn(f'Duplicate senses {sense.id} + {sense2.id} both referring to {sense.synsetid}')


def check_entries(wn: WordnetModel) -> None:
    check_entry_keys(wn)
    for e in wn.entries:
        check_entry_sense_duplicates(e)


# S E N S E S

def check_senseid_duplicates(wn: WordnetModel) -> None:
    visited_sense_ids = {}
    for sense in wn.senses:
        if sense.id in visited_sense_ids:
            warn(f'Duplicate sense id {sense.id}')
        visited_sense_ids[sense.id] = sense.id


def check_sense(wn: WordnetModel, sense: Sense) -> None:
    # sense id is present
    if not sense.id:
        warn("Sense %s does not have a sense id" % sense)
    # id is valid
    check_valid_sense_id(sense.id)
    check_valid_sense_id_for_target(sense.id, sense.synsetid)
    # sensekey is well-formed as expected
    calc_sense_key = make_sense_key(wn, sense.entry, sense)
    if sense.id != calc_sense_key:
        warn(f'Sense {sense.id} should have {calc_sense_key} id')
    # synset reference resolves
    try:
        wn.synset_resolver[sense.synsetid]
    except KeyError as _:
        warn(f'Sense {sense.id} refers to nonexistent synset {sense.synsetid}')


def check_sense_relations(wn: WordnetModel, sense: Sense) -> None:
    synset = wn.synset_resolver[sense.synsetid]
    pos = PartOfSpeech(synset.pos)

    # Iterate
    for r in sense.relations:
        if not r.other_type and Sense.Relation.Type(r.relation_type) == Sense.Relation.Type.PERTAINYM:
            if not equal_pos(pos, PartOfSpeech.ADJECTIVE) and not equal_pos(pos, PartOfSpeech.ADVERB):
                warn(f'Pertainym {r.target} of {sense.id} should be between adjectives')

    # Duplicates
    counter = Counter((r.target, r.relation_type) for r in sense.relations)
    for item, count in counter.items():
        if count > 1:
            warn(f'Duplicate relation {sense.id} ={item[1]}=> {item[0]}')


def check_sense_verbframes(sense: Sense) -> None:
    counter = Counter(sense.verbframeids)
    for item, count in counter.items():
        if count > 1:
            warn(f'Duplicate verb frames in entry {sense.id}')


def check_senses(wn: WordnetModel) -> None:
    check_senseid_duplicates(wn)
    for sense in wn.senses:
        check_sense(wn, sense)
        check_sense_relations(wn, sense)
        check_sense_verbframes(sense)


# S Y N S E T S

def check_members(wn: WordnetModel, synset: Synset) -> None:
    if not synset.members:
        warn(f'Synset {synset.id} members empty')
    try:
        for m in synset.members:
            _ = wn.member_resolver[m, synset.id]
    except KeyError as ke:
        warn(f'Synset {synset.id} refers to nonexistent member {ke.args[0]}')


def check_synset_relations(wn: WordnetModel, synset: Synset) -> None:
    pos = PartOfSpeech(synset.pos)

    # Iterate
    for r in synset.relations:
        t = Synset.Relation.Type(r.relation_type)
        # resolve target
        try:
            target = wn.synset_resolver[r.target]
            # cross-pos hypernym
            if t == Synset.Relation.Type.HYPERNYM and not equal_pos(pos, PartOfSpeech(target.pos)):
                warn(f'Cross-part-of-speech hypernym {synset.id} => {r.target}')
            # no synset antonym
            if t == Synset.Relation.Type.ANTONYM:
                warn(f'Antonymy should be at the sense level {synset.id} => {r.target}')
        except KeyError as _:
            warn(f'{synset.id} refers to nonexistent synset {r.target}')

    # Duplicates
    sorted_relations = sorted(synset.relations, key=lambda _: (_.target, _.relation_type))
    for i in range(len(sorted_relations) - 1):
        r = sorted_relations[i]
        r_next = sorted_relations[i + 1]
        if r.target == r_next.target and Synset.Relation.Type(r.relation_type) == Synset.Relation.Type(r_next.relation_type):
            warn(f'Duplicate synset relation {synset.id} ={r.relation_type}=> {r.target}')
    # TODO redundant
    counter = Counter((r.target, r.relation_type) for r in synset.relations)
    for item, count in counter.items():
        if count > 1:
            warn(f'Duplicate synset relation {synset.id} ={item[1]}=> {item[0]}')

    # Similar/ Adj clusters
    def count_similars(ss: Synset):
        result = 0
        for r2 in ss.relations:
            t2 = Synset.Relation.Type(r2.relation_type)
            if t2 == Synset.Relation.Type.SIMILAR:
                if not equal_pos(pos, PartOfSpeech.VERB) and not equal_pos(pos, PartOfSpeech.ADJECTIVE):
                    warn(f'Similar relation not between verb/adjective {ss.id} => {r2.target}')
                result += 1
                if result > 1 and pos == PartOfSpeech.ADJECTIVE_SATELLITE:
                    warn(f'Satellite of more than one synset {ss.id}')
        return result

    if pos == PartOfSpeech.ADJECTIVE_SATELLITE:
        if count_similars(synset) == 0:
            warn(f'Satellite must have at least one similar relation {synset.id}')

    # Noun hypernyms
    top = "00001740-n"
    if pos == PartOfSpeech.NOUN:
        hypernyms = [r for r in synset.relations if
                     Synset.Relation.Type(r.relation_type) == Synset.Relation.Type.HYPERNYM or
                     Synset.Relation.Type(r.relation_type) == Synset.Relation.Type.INSTANCE_HYPERNYM]
        if not hypernyms and synset.id != top:
            warn(f'Noun synset {synset.id} has no hypernym')


def check_instances(wn: WordnetModel) -> None:
    def collect_instances():
        result = set()
        for ss in wn.synsets:
            if any(Synset.Relation.Type(r2.relation_type) == Synset.Relation.Type.INSTANCE_HYPERNYM for r2 in ss.relations):
                if any(Synset.Relation.Type(r3.relation_type) == Synset.Relation.Type.HYPERNYM for r3 in ss.relations):
                    warn(f'Synset {ss.id} has both hypernym and instance hypernym')
                result.add(ss.id)
        return result

    instances = collect_instances()
    for synset in wn.synsets:
        for r in synset.relations:
            if Synset.Relation.Type(r.relation_type) == Synset.Relation.Type.HYPERNYM:
                if r.target in instances:
                    warn(f'Hypernym targets instance {synset.id} => {r.target}')


def check_ili(synset: Synset) -> None:
    if (not synset.ili or synset.ili == "in") and not synset.ili_definition:
        pass  # TODO print("%s does not have an ILI definition" % ss.id)


def check_synset(wn: WordnetModel, synset: Synset) -> None:
    pos = PartOfSpeech(synset.pos)

    # id
    check_valid_synset_id(synset.id)
    if synset.id[-1:] != pos.value:
        warn(f'Synset ID {synset.id} clashes with part-of-speech {pos.value}')
    # members
    check_members(wn, synset)
    # definitions
    if len(synset.definitions) == 0:
        warn(f'Synset without definition {synset.id}')
    for defn in synset.definitions:
        if len(defn) == 0:
            warn('Synset with empty definition {synset.id}')
    # ili
    check_ili(synset)

    # Synset relations
    check_synset_relations(wn, synset)


def check_synsets(wn: WordnetModel) -> None:
    for synset in wn.synsets:
        check_synset(wn, synset)
    check_instances(wn)


def main(wn: WordnetModel) -> None:
    check_entries(wn)

    check_senses(wn)
    check_synsets(wn)

    check_symmetry(wn)
    check_transitive(wn)
    check_no_loops(wn)
    check_no_domain_loops(wn)


if __name__ == "__main__":
    arg_parser = argparse.ArgumentParser(description="load from yaml and save")
    arg_parser.add_argument('--pickle', action='store_true', default=False, help='use pickle')
    arg_parser.add_argument('in_dir', type=str, help='from-dir for yaml/pickle')
    arg_parser.add_argument('pickled', type=str, nargs='?', default='oewn.pickle', help='from-pickle')
    args = arg_parser.parse_args()


    def get_wn() -> WordnetModel:
        if args.pickle:
            from oewn_core.deserialize import load as pickle_load
            return pickle_load(args.in_dir, args.pickled) #, extend=True
        from oewn_core.wordnet_fromyaml import load as yaml_load
        return yaml_load(args.in_dir) # , extend=True


    _wn: WordnetModel = get_wn()
    try:
        main(_wn)
        print("No validity issues")
    except ValidationError as ve:
        print(ve, file=sys.stderr)



#!/usr/bin/python3

"""
WordNet deserialize

Author: John McCrae <john@mccr.ae> for original code
Author: Bernard Bou <1313ou@gmail.com> for rewrite and revamp
"""

#  Copyright (c) 2024.
#  Creative Commons 4 for original code
#  GPL3 for rewrite

import argparse
import pickle
import sys
import time

from oewn_core.wordnet import WordnetModel


def load_pickle(path: str, file='wn.pickle') -> WordnetModel:
    """
    Load model from pickle file in path
    """
    with open(f"{path}/{file}", "rb") as out:
        return pickle.load(out)


def load(home: str, file='oewn.pickle', extend: bool = True, resolve: bool = False, verbose: bool = False) -> WordnetModel:
    if verbose:
        print(f'loading from pickle {file} in {home}')
    wn = load_pickle(home, file=file)
    if verbose:
        print(f'loaded {wn} from pickle {file} in {home}')
    if extend:
        if verbose:
            print(f'extending relations')
            print(f'before extension: {wn.info_relations()}')
        wn.extend()
        if verbose:
            print(f'after extension:  {wn.info_relations()}')
            print(f'extended relations')
    if resolve:
        if verbose:
            print(f'resolving cross-references')
        wn.resolve()
        if verbose:
            print(f'resolved cross-references')
    if verbose:
        print(wn)
        print(wn.info())
        print(wn.info_relations())
    return wn


def test(wn, out_dir) -> None:
    import oewn_core.wordnet_toyaml as saver
    saver.save(wn, out_dir)


def main() -> WordnetModel:
    """
    WordNet load-save
    Will have a normalizing effect, after which it's not modified
    """
    arg_parser = argparse.ArgumentParser(description="load from yaml and save")
    arg_parser.add_argument('in_dir', type=str, help='from-dir for pickle')
    arg_parser.add_argument('pickled', type=str, nargs='?', default='oewn.pickle', help='from-pickle')
    args = arg_parser.parse_args()
    return load(args.in_dir, args.pickled)


if __name__ == '__main__':
    start_time = time.time()
    main()
    end_time = time.time()
    duration = end_time - start_time
    print(f"Loading took {duration:.6f} seconds", file=sys.stderr)



#!/usr/bin/python3

"""
WordNet load from XML-save to YAML
Will have a normalizing effect, after which it's idempotent

Author: John McCrae <john@mccr.ae> for original code
Author: Bernard Bou <1313ou@gmail.com> for rewrite and revamp
"""

#  Copyright (c) 2024.
#  Creative Commons 4 for original code
#  GPL3 for rewrite

import argparse
import sys
import time

from oewn_core.wordnet_toyaml import save
from oewn_xml.wordnet_fromxml import load


def xml1_to_yaml(in_file, out_dir) -> None:
    wn = load(in_file)
    save(wn, out_dir)


def main() -> None:
    arg_parser = argparse.ArgumentParser(description="load from xml and save to yaml")
    arg_parser.add_argument('in_file', type=str, help='from-file')
    arg_parser.add_argument('out_dir', type=str, help='to-dir')
    args = arg_parser.parse_args()
    xml1_to_yaml(args.in_file, args.out_dir)


if __name__ == '__main__':
    start_time = time.time()
    main()
    end_time = time.time()
    duration = end_time - start_time
    print(f"XML-YAML conversion took {duration:.6f} seconds", file=sys.stderr)



"""
WordNet XML common

Author: John McCrae <john@mccr.ae> for original code
Author: Bernard Bou <1313ou@gmail.com> for rewrite and revamp
Author: Michael Wayne Goodman <goodman.m.w@gmail.com> for escaping
"""
#  Copyright (c) 2024.
#  Creative Commons 4 for original code
#  GPL3 for rewrite

import re
from abc import abstractmethod, ABC
from typing import Any, Dict, Tuple

# Constrain input to avoid non letters or unescaped chars, or not
unconstrained = False

# X M L   I D

# Regular expressions for valid NameStartChar and NameChar based on the XML 1.1 specification.
# based on the XML 1.0 specification.
# We don't check for 1st character extra restrictions
# because it's always prefixed with 'oewn-'
xml_id_punct = r':_'  # colon, underscore
xml_id_punct_not_first = r'\-\.·'  # hyphen-minus, period, middle dot
xml_id_az = r'A-Za-z'
xml_id_num = r'0-9'
xml_id_extend = (
    r'\xC0-\xD6'  # Latin letters with diacritics
    r'\xD8-\xF6'  # Additional latin letters with diacritics
    r'\xF8-\u02FF'  # Extended Latin letters and characters from the Latin Extended-A, Greek, and other blocks
    r'\u0370-\u037D'  # Greek letters and many characters from scripts such as Coptic, Armenian, Hebrew, Arabic, and more
    r'\u037F-\u1FFF'  # Greek letters and many characters from scripts such as Coptic, Armenian, Hebrew, Arabic, and more
    r'\u200C-\u200D'  # Zero Width Non-Joiner (ZWNJ) and Zero Width Joiner (ZWJ), used in scripts like Arabic and Indic.
    r'\u2070-\u218F'  # Superscripts, subscripts, and various other symbols.
    r'\u2C00-\u2FEF'  # Characters from the Glagolitic and other historical scripts.
    r'\u3001-\uD7FF'  # Characters from many Asian scripts, including Chinese, Japanese, and Korean ideographs
    r'\uF900-\uFDCF'  # Compatibility ideographs and Arabic presentation forms
    r'\uFDF0-\uFFFD'  # Compatibility ideographs and Arabic presentation forms
    r'\U00010000-\U000EFFFF'
    # Supplementary chars from planes outside the Basic Multilingual Plane, including rare historical scripts, musical symbols, emoji, and more.
)
xml_id_extend_not_first = (
    r'\u0300-\u036F'  # Combining diacritical marks that can modify the preceding character.
    r'\u203F-\u2040'  # Characters like the "undertie" (‿) and "character tie" (⁀), used in specialized phonetic notations.
)
xml_id_start_char = fr'[{xml_id_punct}{xml_id_az}{xml_id_extend}]'  # not used if oewn- prefix
xml_id_char = fr'[{xml_id_punct}{xml_id_punct_not_first}{xml_id_az}{xml_id_num}{xml_id_extend}{xml_id_extend_not_first}]'

xml_id_start_char1 = fr'^{xml_id_start_char}$'
xml_id_char1 = fr'^{xml_id_char}$'

xml_id_start_char1_re: re.Pattern[str] = re.compile(xml_id_start_char1)
xml_id_char1_re: re.Pattern = re.compile(xml_id_char1)

xml_id = fr'^{xml_id_start_char}{xml_id_char}*$'
xml_id_re: re.Pattern = re.compile(xml_id)


# V A L I D I T Y

def is_valid_xml_id_char(c) -> bool:
    return bool(xml_id_char1_re.match(c))


def is_valid_xml_id(s) -> bool:
    return bool(xml_id_re.match(s))


# E S C A P I N G

def split_at_last(s, c) -> Tuple[Any, Any]:
    cut = s.rfind(c)
    if cut > -1:
        return s[:cut], s[cut + len(c):]
    raise ValueError(f'"{s}" to be cut at last "{c}"')


class NameFactory(ABC):
    """
    Abstract Base Class / Interface
    """

    @abstractmethod
    def escape_lemma(self, lemma: str) -> str:
        pass

    @abstractmethod
    def unescape_lemma(self, esc_lemma: str) -> str:
        pass

    @abstractmethod
    def escape_sensekey(self, sensekey: str) -> str:
        pass

    @abstractmethod
    def unescape_sensekey(self, esc_sensekey: str) -> str:
        pass


class LegacyNameFactory(NameFactory):
    def escape_lemma(self, lemma) -> str:
        """Escape the lemma so that it contains valid characters for inclusion in XML ID"""

        def escape_char(c):
            if ('A' <= c <= 'Z') or ('a' <= c <= 'z') or ('0' <= c <= '9') or c == '.':
                return c
            elif c == ' ':
                return '_'
            elif c == '(':
                return '-lb-'
            elif c == ')':
                return '-rb-'
            elif c == '\'':
                return '-ap-'
            elif c == '/':
                return '-sl-'
            elif c == ':':
                return '-cn-'
            elif c == ',':
                return '-cm-'
            elif c == '!':
                return '-ex-'
            elif c == '+':
                return '-pl-'
            elif xml_id_char1_re.match(c):  # or xml_id_start_char1_re.match(c):
                return c
            raise ValueError(f'Illegal character {c}')

        return "".join(escape_char(c) for c in lemma)

    def unescape_lemma(self, esc_lemma):
        """Reverse escaping the lemma back to the WN lemma"""
        return (esc_lemma
                .replace('-pl-', '+')
                .replace('-ex-', '!')
                .replace('-cm-', ',')
                .replace('-cn-', ':')
                .replace('-sl-', '/')
                .replace('-ap-', "'")
                .replace('-rb-', ')')
                .replace('-lb-', '(')
                .replace('_', ' '))

    xml_percent_sep = '__'
    xml_colon_sep = '.'

    def escape_sensekey(self, sensekey) -> str:
        """
        Maps a sense key into an OEWN form
        """

        def escape_lemma_in_sensekey(lemma):
            return (lemma
                    .replace("'", "-ap-")
                    .replace("/", "-sl-")
                    .replace("!", "-ex-")
                    .replace(",", "-cm-")
                    .replace(":", "-cn-")
                    .replace("+", "-pl-"))

        if '%' in sensekey:
            l, lex_sense = split_at_last(sensekey, '%')
            esc_lemma = escape_lemma_in_sensekey(l)
            lex_sense_fields = lex_sense.split(':')
            n = len(lex_sense_fields)
            assert n == 5, f'Parsing error: length {n} of lex_sense_fields in lex_sense {lex_sense} should be 5'
            head = lex_sense_fields[3]
            if head:
                lex_sense_fields[3] = escape_lemma_in_sensekey(head)
            return f"{esc_lemma}{self.xml_percent_sep}{self.xml_colon_sep.join(lex_sense_fields)}"
        raise ValueError(f'Ill-formed OEWN sense key (no %): {sensekey}')

    def unescape_sensekey(self, esc_sensekey) -> str:
        """
        Maps an OEWN sense key to a WN sense key
        """

        def unescape_lemma_in_sensekey(esc_lemma):
            return (esc_lemma
                    .replace("-ap-", "'")
                    .replace("-sl-", "/")
                    .replace("-ex-", "!")
                    .replace("-cm-", ",")
                    .replace("-cn-", ":")
                    .replace("-pl-", "+"))

        if self.xml_percent_sep in esc_sensekey:
            l, lex_sense = split_at_last(esc_sensekey, self.xml_percent_sep)
            lemma = unescape_lemma_in_sensekey(l)
            lex_sense_fields = lex_sense.split(self.xml_colon_sep)
            n = len(lex_sense_fields)
            assert n == 5, f'Parsing error: length {n} of lex_sense_fields in lex_sense {lex_sense} should be 5'
            head = lex_sense_fields[3]
            if head:
                lex_sense_fields[3] = unescape_lemma_in_sensekey(head)
            return f'{lemma}%{':'.join(lex_sense_fields)}'
        raise ValueError(f'Ill-formed OEWN sense key (no {self.xml_percent_sep}): {esc_sensekey}')


class DashNameFactory(NameFactory):
    esc_char_escapes: Dict[str, str] = {
        '-': '--',  # custom
    }
    base_char_escapes: Dict[str, str] = {
        # HTML entities
        # https://en.wikipedia.org/wiki/List_of_XML_and_HTML_character_entity_references
        "'": '-apos-',
        '`': '-grave-',
        '´': '-acute-',
        '‘': '-lsquo-',
        '’': '-rsquo-',
        '(': '-lpar-',
        ')': '-rpar-',
        '[': '-lsqb-',
        ']': '-rsqb-',
        '{': '-lbrace-',
        '}': '-rbrace-',
        ',': '-comma-',
        ';': '-semi-',
        '=': '-equals-',
        '+': '-plus-',
        '!': '-excl-',
        '?': '-quest-',
        '@': '-commat-',
        '#': '-num-',
        '%': '-percnt-',
        '&': '-amp-',
        '§': '-sect-',
        '¶': '-para-',
        '/': '-sol-',
        '\\': '-bsol-',
        '|': '-vert-',
        '^': '-Hat-',
        '*': '-ast-',
        '$': '-dollar-',
        '¢': '-cent-',
        '£': '-pound-',
        '©': '-copy-',
        '®': '-reg-',
        'º': '-ordm-',
        '°': '-deg-',
        '~': '-tilde-',
    }
    extra_char_escapes: Dict[str, str] = {
        '_': '-lowbar-',
        ' ': '_',
        ':': '-colon-',  # to make valid xsd:ids otherwise not necessary for XML:IDs
    }
    sk_char_escapes: Dict[str, str] = {
        '-': '--',
        ':': '-colon-',  # to make valid xsd:ids otherwise not necessary for XML:IDs
    }

    char_escapes: Dict[str, str] = esc_char_escapes | base_char_escapes | extra_char_escapes
    char_escapes_reverse: Dict[str, str] = {v: k for k, v in char_escapes.items()}

    char_escapes_for_sk: Dict[str, str] = base_char_escapes | sk_char_escapes
    char_escapes_for_sk_reverse: Dict[str, str] = {v: k for k, v in char_escapes_for_sk.items()}

    def __init__(self, main_separator, minor_separator) -> None:
        self.xml_percent_sep = main_separator
        self.xml_colon_sep = minor_separator

    # l e m m a

    def escape_lemma(self, lemma) -> str:
        """
        Escape the lemma so that it contains only valid characters for inclusion in XML ID
        """

        def escape_char(c):
            if ('A' <= c <= 'Z') or ('a' <= c <= 'z') or ('0' <= c <= '9'):
                return c
            elif c in self.char_escapes:
                return self.char_escapes[c]
            elif xml_id_char1_re.match(c):
                return c
            elif unconstrained:
                return f'-{ord(c):04X}-'
            raise ValueError(f'{c!r} [x{ord(c):04X}] is illegal character in XML ID and no escape sequence is defined')

        return ''.join(escape_char(c) for c in lemma)

    def unescape_lemma(self, esc_lemma):
        """
        Reverse the escaping and retrieve the original lemma
        Reversing the keys matters because application order matters
        """

        s = esc_lemma
        for seq in reversed(self.char_escapes_reverse.keys()):
            s = s.replace(seq, self.char_escapes_reverse[seq])
        return s

    # s e n s e k e y

    def escape_lemma_in_sensekey(self, lemma) -> str:
        """
        Escape the lemma so that it contains only valid characters for inclusion in XML ID
        within the context of sense id factory
        """

        def escape_char(c):
            if ('A' <= c <= 'Z') or ('a' <= c <= 'z') or ('0' <= c <= '9'):
                return c
            elif c in self.char_escapes_for_sk:
                return self.char_escapes_for_sk[c]
            elif xml_id_char1_re.match(c):
                return c
            elif unconstrained:
                return f'-{ord(c):04X}-'
            raise ValueError(f'{c!r} [x{ord(c):04X}] is illegal character in XML ID and no escape sequence is defined')

        return ''.join(escape_char(c) for c in lemma)

    def unescape_lemma_in_sensekey(self, esc_lemma):
        """
        Reverse the escaping and retrieve the original lemma
        within the context of sense id factory
        """
        s = esc_lemma
        for seq in self.char_escapes_for_sk_reverse:
            s = s.replace(seq, self.char_escapes_for_sk_reverse[seq])
        return s

    def escape_sensekey(self, sensekey) -> str:
        """Escape the sensekey so that it contains valid characters for XML ID"""
        if '%' in sensekey:
            l, lex_sense = split_at_last(sensekey, '%')
            lemma = self.escape_lemma_in_sensekey(l)
            lex_sense_fields = lex_sense.split(':')
            n = len(lex_sense_fields)
            assert n == 5, f'Parsing error: length {n} of lex_sense_fields in lex_sense {lex_sense} should be 5'
            head = lex_sense_fields[3]
            if head:
                lex_sense_fields[3] = self.escape_lemma_in_sensekey(head)
            return f"{lemma}{self.xml_percent_sep}{self.xml_colon_sep.join(lex_sense_fields)}"
        raise ValueError(f'Ill-formed OEWN sense key (no %): {sensekey}')

    def unescape_sensekey(self, esc_sensekey) -> str:
        """
        Unescape an OEWN sense key to a WN sense key
        """
        if self.xml_percent_sep in esc_sensekey:
            l, lex_sense = split_at_last(esc_sensekey, self.xml_percent_sep)
            lemma = self.unescape_lemma_in_sensekey(l)
            lex_sense_fields = lex_sense.split(self.xml_colon_sep)
            n = len(lex_sense_fields)
            assert n == 5, f'Parsing error: length {n} of lex_sense_fields in lex_sense {lex_sense} should be 5'
            head = lex_sense_fields[3]
            if head:
                lex_sense_fields[3] = self.unescape_lemma_in_sensekey(head)
            return f'{lemma}%{':'.join(lex_sense_fields)}'
        raise ValueError(f'Ill-formed OEWN sense key (no {self.xml_percent_sep}): {esc_sensekey}')


# S Y N S E T / E N T R Y  /  S E N S E   X M L   I D

# Name factory

legacy_factory: LegacyNameFactory = LegacyNameFactory()
middledot = '·'
dash_factory: DashNameFactory = DashNameFactory(middledot, '.')
default_factory: NameFactory = dash_factory


def escape_lemma(lemma) -> str:
    return default_factory.escape_lemma(lemma)


def unescape_lemma(esc_lemma) -> str:
    return default_factory.unescape_lemma(esc_lemma)


def escape_sensekey(sensekey) -> str:
    return default_factory.escape_sensekey(sensekey)


def unescape_sensekey(esc_sensekey) -> str:
    return default_factory.unescape_sensekey(esc_sensekey)


# Prefix to XML IDs

key_prefix = 'oewn-'
key_prefix_len = len(key_prefix)


def is_valid_xml_oewn_id(_):
    return _.startswith(key_prefix) and is_valid_xml_id(_)


# s y n s e t

def to_xml_synset_id(synsetid: str) -> str:
    return f'{key_prefix}{synsetid}'


def from_xml_synset_id(xml_synsetid: str) -> str:
    return xml_synsetid[key_prefix_len:]


# e n t r y

def to_xml_entry_id(lemma: str, pos: str, discriminant: str | None = None, name_factory=default_factory):
    p = pos
    d = f'-{discriminant}' if discriminant else ''
    return f'{key_prefix}{name_factory.escape_lemma(lemma)}-{p}{d}'


def from_xml_entry_id(xml_entry_id: str, name_factory=default_factory) -> Tuple[str, str, str | None]:
    entry_id2 = xml_entry_id[key_prefix_len:]
    pos_discriminant = entry_id2[-1]
    if pos_discriminant in ('n', 'v', 'a', 'r'):
        lemma = entry_id2[:-2]
        pos = pos_discriminant
        discriminant = None
    else:
        def penultimate_occurrence(s, char):
            last_index = s.rfind(char)
            if last_index == -1:  # Character not found
                return -1
            return s.rfind(char, 0, last_index)

        cut = penultimate_occurrence(entry_id2, '-')
        if cut == -1:
            raise ValueError(xml_entry_id)
        lemma = entry_id2[:cut]
        pos_discriminant = entry_id2[cut + 1:]
        f = pos_discriminant.split('-')
        pos = f[0]
        discriminant = f[1]
    lemma = name_factory.unescape_lemma(lemma)
    return lemma, pos, discriminant


# s e n s e

def to_xml_sense_id(sensekey, name_factory=default_factory) -> str:
    """
    Maps the sensekey so that it contains valid characters for XML ID, prefix added
    :param sensekey: sense key in WN format (YAML)
    :param name_factory: name factory used
    :return: sense key in XML format
    """
    return f"{key_prefix}{name_factory.escape_sensekey(sensekey)}"


def from_xml_sense_id(xml_sensekey, name_factory=default_factory):
    """
    Maps an OEWN XML sense key back to a WN one
    :param xml_sensekey: sense key in XML format
    :param name_factory: name factory used
    :return: sense key in WN format (YAML)
    """
    return name_factory.unescape_sensekey(xml_sensekey[key_prefix_len:])


# L I T E R A L S

def escape_xml_lit(lit):
    """
    Escape the literal for XML
    :param lit: literal
    :return: escaped literal with escaped entities
    """
    return (lit
            .replace("&", "&amp;")
            .replace("'", "&apos;")
            .replace("\"", "&quot;")
            .replace("<", "&lt;")
            .replace(">", "&gt;"))



"""
WordNet model made available for tests
Author: John McCrae <john@mccr.ae> for original code
Author: Bernard Bou <1313ou@gmail.com> for rewrite and revamp
"""
#  Copyright (c) 2024.
#  Creative Commons 4 for original code
#  GPL3 for rewrite

import os
import sys
from typing import List

from oewn_core.deserialize import load
from oewn_core.wordnet import WordnetModel, Sense, Entry

data_home = os.environ['OEWN_HOME']
print(f'data={data_home}', file=sys.stderr)
wn: WordnetModel = load(data_home, extend=False)

sorted_entries: List[Entry] = sorted(list(wn.entries), key=lambda e: e.lemma)
sorted_senses: List[Sense] = sorted(list(wn.senses), key=lambda s: s.id)



"""
WordNet to-XML utilities

Author: John McCrae <john@mccr.ae> for original code
Author: Bernard Bou <1313ou@gmail.com> for rewrite and revamp
"""

#  Copyright (c) 2024.
#  Creative Commons 4 for original code
#  GPL3 for rewrite

import uuid
from datetime import datetime
from typing import List, Dict, Tuple

from oewn_core.wordnet import WordnetModel, Sense, Entry, Synset, Pronunciation, Example, VerbFrame
from oewn_xml.wordnet_xml import to_xml_sense_id, to_xml_synset_id, escape_xml_lit, to_xml_entry_id

I = '  '  # indentation


def lexicon_to_xml(wn: WordnetModel, out, comments=None) -> None:
    """
    Lexicon to XML
    :param wn: lexicon
    :param out: output XML file
    :param comments: optional comments for relations
    :return: None
    """
    out.write('<?xml version="1.0" encoding="UTF-8"?>\n')
    out.write('<!DOCTYPE LexicalResource SYSTEM "http://globalwordnet.github.io/schemas/WN-LMF-1.1.dtd">\n')
    out.write('<LexicalResource xmlns:dc="https://globalwordnet.github.io/schemas/dc/">\n')
    out.write(f"""{I * 1}<Lexicon id="{wn.id}"
        label="{wn.label}"
        language="{wn.language}"
        email="{wn.email}"
        license="{wn.license}"
        version="{wn.version}"
        url="{wn.url}"
        dc:identifier="{uuid.uuid4()}"
        dc:date="{datetime.now().strftime("%Y-%m-%d %H:%M")}">\n""")
    for entry in sorted(wn.entries, key=lambda e: make_entry_id_from_entry(e)):  # key=lambda e: e.lemma):
        entry_to_xml(entry, out, comments)
    for synset in sorted(wn.synsets, key=lambda ss: ss.id):
        synset_to_xml(synset, wn.member_resolver, out, comments)
    for verbframe in sorted(wn.verbframes, key=lambda f: f.id):
        verbframe_to_xml(verbframe, out)
    out.write(f'{I * 1}</Lexicon>')
    out.write('</LexicalResource>\n')


def entry_to_xml(entry: Entry, out, comments) -> None:
    eid = make_entry_id_from_entry(entry)
    out.write(f'{I * 2}<LexicalEntry id="{eid}">\n')
    lemma_to_xml(entry.lemma, entry.pos, entry.pronunciations, out)
    for form in entry.forms:
        form_to_xml(form, out)
    for sense in entry.senses:
        sense_to_xml(sense, out, comments)
    out.write(f'{I * 2}</LexicalEntry>\n')


def sense_to_xml(sense: Sense, out, comments) -> None:
    a = f' adjposition="{sense.adjposition}"' if sense.adjposition else ''
    c = f' subcat="{' '.join(sense.verbframeids)}"' if sense.verbframeids else ''
    n = ''  # f' n="sense.n"'
    sid = to_xml_sense_id(sense.id)
    ssid = to_xml_synset_id(sense.synsetid)
    if len(sense.relations) > 0 or len(sense.examples) > 0:
        out.write(f'{I * 3}<Sense id="{sid}"{n}{a}{c} synset="{ssid}">\n')
        for rel in sense.relations:
            sense_relation_to_xml(rel, out, comments)
        for ex in sense.examples:
            example_to_xml(ex, 4, out)
        out.write(f'{I * 3}</Sense>\n')
    else:
        out.write(f'{I * 3}<Sense id="{sid}"{n}{a}{c} synset="{ssid}"/>\n')


def synset_to_xml(synset: Synset, member_resolver: Dict[Tuple[str, str], Entry], out, comments):
    if comments and synset.id in comments:
        out.write(f'{I * 2}<!-- %s -->\n' % comments[synset.id])
    ssid = to_xml_synset_id(synset.id)
    p = synset.pos
    m = ' '.join([make_entry_id_from_member(m, synset.id, member_resolver) for m in synset.members])
    l = synset.lex_name
    i = synset.ili
    s = f' dc:source="{synset.source}"' if synset.source else ''
    w = f' dc:subject="{synset.wikidata}"' if synset.wikidata else ''
    out.write(f'{I * 2}<Synset id="{ssid}" ili="{i}" members="{m}" partOfSpeech="{p}" lexfile="{l}"{s}{w}>\n')

    for definition in synset.definitions:
        definition_to_xml(definition, out)
    if synset.ili_definition:
        definition_to_xml(synset.ili_definition, out, True)
    for rel in synset.relations:
        synset_relation_to_xml(rel, out, comments)
    for ex in synset.examples:
        example_to_xml(ex, 3, out)
    for us in synset.usages:
        usage_to_xml(us, out)
    out.write(f'{I * 2}</Synset>\n')


def lemma_to_xml(lemma: str, pos: str, pronunciations: List[Pronunciation], out):
    w = escape_xml_lit(lemma)
    p = pos
    if pronunciations:
        out.write(f'{I * 3}<Lemma writtenForm="{w}" partOfSpeech="{p}">\n')
        for pronunciation in pronunciations:
            pronunciation_to_xml(pronunciation, out)
        out.write(f'{I * 3}</Lemma>\n')
    else:
        out.write(f'{I * 3}<Lemma writtenForm="{w}" partOfSpeech="{p}"/>\n')


def pronunciation_to_xml(pronunciation: Pronunciation, out) -> None:
    v = f' variety="{pronunciation.variety}"' if pronunciation.variety else ''
    p = escape_xml_lit(pronunciation.value)
    out.write(f'{I * 4}<Pronunciation{v}>{p}</Pronunciation>\n')


def form_to_xml(form: str, out) -> None:
    f = escape_xml_lit(form)
    out.write(f'{I * 4}<Form writtenForm="{f}"/>\n')


def definition_to_xml(definition: str, out, is_ili=False) -> None:
    d = escape_xml_lit(definition)
    result = f'{I * 3}<ILIDefinition>{d}</ILIDefinition>\n' if is_ili else f'{I * 3}<Definition>{d}</Definition>\n'
    out.write(result)


def example_to_xml(example: str | Example, indent: int, out):
    is_example = isinstance(example, Example)
    e = escape_xml_lit(example.text if is_example else example)
    s = example.source if is_example and example.source else None
    result = f'{I * indent}<Example dc:source="{escape_xml_lit(s)}">{e}</Example>\n' if s else f'{I * indent}<Example>{e}</Example>\n'
    out.write(result)


def usage_to_xml(usage: str, out) -> None:
    u = escape_xml_lit(usage)
    out.write(f'{I * 3}<Usage>{u}</Usage>\n')


def synset_relation_to_xml(synset_relation: Synset.Relation, out, comments):
    r = synset_relation.relation_type
    t = to_xml_synset_id(synset_relation.target)
    out.write(f'{I * 3}<SynsetRelation relType="{r}" target="{t}"/>')
    if comments and synset_relation.target in comments:
        out.write(f' <!-- {comments[synset_relation.target]} -->')
    out.write('\n')


def sense_relation_to_xml(sense_relation: Sense.Relation, out, comments):
    t = to_xml_sense_id(sense_relation.target)
    r = sense_relation.relation_type
    o = sense_relation.other_type
    if o:
        out.write(f'{I * 4}<SenseRelation relType="other" target="{t}" dc:type="{r}"/>')
    else:
        out.write(f'{I * 4}<SenseRelation relType="{r}" target="{t}"/>')
    if comments and sense_relation.target in comments:
        out.write(f' <!-- {comments[sense_relation.target]} -->')
    out.write('\n')


def verbframe_to_xml(verbframe: VerbFrame, out) -> None:
    fid = verbframe.id
    f = escape_xml_lit(verbframe.verbframe)
    out.write(f'{I * 2}<SyntacticBehaviour id="{fid}" subcategorizationFrame="{f}"/>\n')


def make_entry_id_from_entry(entry) -> str:
    return to_xml_entry_id(entry.lemma, entry.pos, entry.discriminant)


def make_entry_id_from_member(lemma: str, synsetid: str, member_resolver: Dict[Tuple[str, str], Entry]):
    if not member_resolver:
        raise ValueError('Null or empty member resolver')
    k = (lemma, synsetid)
    if k not in member_resolver:
        raise ValueError(f'Member resolver cannot resolve {k}')
    e = member_resolver[k]
    return to_xml_entry_id(lemma, e.pos, e.discriminant)


def save(wn: WordnetModel, path) -> None:
    print(f'saving to XML {path}')
    with open(path, 'w', encoding='utf-8') as out:
        lexicon_to_xml(wn, out)
    print(f'saved to XML {path}')



"""
Testing args
"""
#  Copyright (c) 2024.
#  Creative Commons 4 for original code
#  GPL3 for rewrite

import argparse


def list_of(*args):
    return list(args)


def main() -> None:
    arg_parser = argparse.ArgumentParser(description="load wn and syntagnet from yaml, merge and save")
    arg_parser.add_argument('arg1', type=str, help='required/positional')
    arg_parser.add_argument('arg2', type=str, nargs='?', default=None, help='optional/positional')
    arg_parser.add_argument('--opt', action=argparse.BooleanOptionalAction, help='switch')
    arg_parser.add_argument('--opt2', action='store_true', default=False, help='model to use')
    arg_parser.add_argument('--opt3', default=None, help='model to use')

    for cl in (
            list_of('A1', 'A2'),
            list_of('A1', '--opt'),
            list_of('--opt', 'A1'),
            list_of('A1', '--opt', '--opt2'),
            list_of('A1', '--opt', '--opt2', '--opt3', 'extra'),
    ):
        print(cl)
        args = arg_parser.parse_args(cl)
        print(args)


if __name__ == '__main__':
    main()



"""
WordNet XML ID char validity tests
Author: John McCrae <john@mccr.ae> for original code
Author: Bernard Bou <1313ou@gmail.com> for rewrite and revamp
"""
#  Copyright (c) 2024.
#  Creative Commons 4 for original code
#  GPL3 for rewrite

import unittest
from typing import Any, Generator, Tuple

from oewn_xml.wordnet_xml import dash_factory, is_valid_xml_id_char, xml_id_start_char1_re, xml_id_char1_re


def char_range(c1, c2) -> Generator[str, Any, None]:
    """Generates the characters from `c1` to `c2`, inclusive."""
    for c in range(ord(c1), ord(c2) + 1):
        yield chr(c)


def char_ranges(r, ) -> Generator[str, Any, None]:
    for r1, r2 in r:
        for c in char_range(r1, r2):
            yield c


ascii_but_alphanum: Tuple[Tuple[str, str], ...] = (
    ('\x00', '\x2F'),
    ('\x3A', '\x3A'),
    ('\x5F', '\x5F'),
    ('\x7B', '\xBF'),
    # ('\xC0', '\xFF'),
)

printable_ascii_but_alphanum: Tuple[Tuple[str, str], ...] = (
    ('\x20', '\x2F'),
    ('\x3A', '\x3A'),
    ('\x5F', '\x5F'),
    ('\x7B', '\xBF'),
    # ('\xC0', '\xFF'),
)


class XMLIDCharsTestCase(unittest.TestCase):
    def test_ascii_non_letter_range(self) -> None:
        print('\nXML VALID IN ASCII RANGE')
        for c in char_ranges(ascii_but_alphanum):
            start = xml_id_start_char1_re.match(c) is not None
            mid = xml_id_char1_re.match(c) is not None
            if start or mid:
                print(f'{c!r} {ord(c):X} {" start" if start else ''} {"mid" if mid else ''}')
        self.assertTrue(True)

    def test_ascii_invalid_range(self) -> None:
        print('\nESCAPABILITY OF XML INVALID IN ASCII RANGE')
        for c in char_ranges(printable_ascii_but_alphanum):
            if not is_valid_xml_id_char(c):
                is_escaped = c in dash_factory.char_escapes
                print(f'Invalid {c} {c!r} {ord(c):X} {f'ESCAPED AS {dash_factory.char_escapes[c]}' if is_escaped else ''}')
        self.assertTrue(True)

    def test_ascii_diacritics_range(self) -> None:
        print('\nWITH DIACRITICS')
        for c in 'ñéèàâç汉':
            start = is_valid_xml_id_char(c) is not None
            mid = xml_id_char1_re.match(c) is not None
            print(f'{c} {ord(c):X} {" start" if start else ""} {"mid" if mid else ""}')
            self.assertTrue(is_valid_xml_id_char(c))
        self.assertTrue(True)


if __name__ == '__main__':
    unittest.main()



"""
WordNet model browse sample
Author: John McCrae <john@mccr.ae> for original code
Author: Bernard Bou <1313ou@gmail.com> for rewrite and revamp
"""
#  Copyright (c) 2024.
#  Creative Commons 4 for original code
#  GPL3 for rewrite

import argparse
import sys
from glob import glob
from typing import Dict

import yaml


def merge_yaml_files(output_file, files):
    merged_data = {}
    for file in files:
        with open(file) as f:
            data = yaml.safe_load(f)
            if isinstance(data, dict):
                _deep_merge(merged_data, data)
            else:
                print("Cannot merge non dictionaries", file=sys.stderr)
    with open(output_file, 'w') as out:
        yaml.dump(merged_data, out, allow_unicode=True)


def _shallow_merge(target: Dict, source: Dict) -> Dict:
    target.update(source)
    return target


def _deep_merge(base: Dict, incoming: Dict) -> Dict:
    for key, value in incoming.items():
        if isinstance(value, dict) and key in base and isinstance(base[key], dict):
            _deep_merge(base[key], value)
        else:
            base[key] = value
    return base


def main() -> None:
    arg_parser = argparse.ArgumentParser(description="browse")
    arg_parser.add_argument('in_dir', type=str, help='from-dir for yaml/pickle')
    arg_parser.add_argument('out_dir', type=str, help='out-dir for result')
    arg_parser.add_argument('syntagnet', type=str, nargs='?', default='syntagnet.yaml', help='syntagnet data')
    args = arg_parser.parse_args()

    merge_files = glob(f'{args.in_dir}/*.yaml')
    merge_files.append(args.syntagnet)
    merge_yaml_files(f'{args.out_dir}/merged.yaml', merge_files)


if __name__ == '__main__':
    main()



"""
WordNet tests for escapable sequences found in OEWN
Author: John McCrae <john@mccr.ae> for original code
Author: Bernard Bou <1313ou@gmail.com> for rewrite and revamp
"""
#  Copyright (c) 2024.
#  Creative Commons 4 for original code
#  GPL3 for rewrite

import unittest

from oewn_xml.wordnet_xml import DashNameFactory, is_valid_xml_id
from tests.model import wn
from tests.utils import collect_entries_for_escapes, print_as_dictionary, generate_senseid_from_sensekey, is_parsable_xml_sensekey, is_valid_xsd_id


class EscapablesTestCase(unittest.TestCase):
    limit = 5

    def test_escapables(self) -> None:
        r = collect_entries_for_escapes(wn.entries, DashNameFactory.char_escapes_for_sk)
        print_as_dictionary(r)
        for k in r:
            v = r[k]
            if v:
                print(f'{k}')
                for i, e in enumerate(r[k]):
                    for s in e.senses:
                        try:
                            sk, esc_sk, unesc_sk = generate_senseid_from_sensekey(s.id)
                            self.assertTrue(is_parsable_xml_sensekey(esc_sk), f'Unparsable {esc_sk}')
                            self.assertTrue(is_valid_xml_id(esc_sk), f'Invalid XML ID {esc_sk}')
                            self.assertTrue(is_valid_xsd_id(esc_sk), f'Invalid xsd:id {esc_sk}')
                            if i < self.limit:
                                print(f'\t{sk} --xml-->  {esc_sk} --reverse-->  {unesc_sk}')
                        except ValueError as ve:
                            self.assertFalse(False, ve)


if __name__ == '__main__':
    unittest.main()



#  Copyright (c) 2024.
#  Creative Commons 4 for original code
#  GPL3 for rewrite

from setuptools import setup, find_packages

setup(
    name='oewn-core',
    version='1.3.0',
    author='Bernard Bou, John McCrae',
    author_email='oewntk@gmail.com',
    description='Core library for oewn',
    long_description=open('PKG-INFO.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/oewntk/oewn-core',
    packages=find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.9',
    install_requires=[
        'PyYAML',
    ],
    extras_require={
        'dev': [
            'unittest',
        ],
    },
)


"""
WordNet XML ID discriminant handling tests
Author: John McCrae <john@mccr.ae> for original code
Author: Bernard Bou <1313ou@gmail.com> for rewrite and revamp
"""
#  Copyright (c) 2024.
#  Creative Commons 4 for original code
#  GPL3 for rewrite

import unittest

from oewn_xml.wordnet_xml import from_xml_entry_id, to_xml_entry_id


class DiscriminantTestCase(unittest.TestCase):

    def test_discriminant(self) -> None:
        print('\nDISCRIMINANT')
        d1 = '1'
        d2 = '2'
        for lemma in ('bass', 'bow', 'row', 'wind'):
            e_id_0 = to_xml_entry_id(lemma, 'n')
            e_id_1 = to_xml_entry_id(lemma, 'n', d1)
            e_id_2 = to_xml_entry_id(lemma, 'n', d2)
            l_0, p_0, d_0 = from_xml_entry_id(e_id_0)
            l_1, p_1, d_1 = from_xml_entry_id(e_id_1)
            l_2, p_2, d_2 = from_xml_entry_id(e_id_2)
            print(f'{lemma} --entry()--> {e_id_0} --parse--> {l_0} {p_0} {d_0}')
            print(f'{lemma} --entry(1)--> {e_id_1} --parse--> {l_1} {p_1} {d_1}')
            print(f'{lemma} --entry(2)--> {e_id_2} --parse--> {l_2} {p_2} {d_2}')

            self.assertEqual(l_0, lemma)
            self.assertEqual(l_1, lemma)
            self.assertEqual(l_2, lemma)
            self.assertEqual(p_0, 'n')
            self.assertEqual(p_1, 'n')
            self.assertEqual(p_2, 'n')
            self.assertEqual(d_0, None)
            self.assertEqual(d_1, d1)
            self.assertEqual(d_2, d2)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python3

"""
WordNet persistence

Author: John McCrae <john@mccr.ae> for original code
Author: Bernard Bou <1313ou@gmail.com> for rewrite and revamp
"""

#  Copyright (c) 2024.
#  Creative Commons 4 for original code
#  GPL3 for rewrite

import argparse
import glob
import sys
import time

import yaml


def normalize(home: str, verbose: bool = False) -> None:
    """ Normalize (home dir)"""
    if verbose:
        print(f'normalizing YAML in {home}')
    for f in glob.glob(f"{home}/*.yaml"):
        data = yaml.load(open(f), Loader=yaml.CLoader)
        with open(f, "w", encoding='utf-8') as out:
            yaml.dump(data, out, allow_unicode=True)
    if verbose:
        print(f'normalized YAML in {home}')


def main() -> None:
    """
    WordNet normalize
    Will test YAML well-formedness
    Will have a limited normalizing effect, after which it's not modified
    YAML keys are not alphabetically reordered,
    turn to load()-save() if this is desired
    """
    arg_parser = argparse.ArgumentParser(description="load from yaml and save")
    arg_parser.add_argument('in_dir', type=str, help='from-dir')
    args = arg_parser.parse_args()
    normalize(args.in_dir)

if __name__ == '__main__':
    start_time = time.time()
    main()
    end_time = time.time()
    duration = end_time - start_time
    print(f"Normalizing took {duration:.6f} seconds", file=sys.stderr)



#!/usr/bin/python3

"""
WordNet from-YAML utilities

Author: John McCrae <john@mccr.ae> for original code
Author: Bernard Bou <1313ou@gmail.com> for rewrite and revamp
"""
#  Copyright (c) 2024.
#  Creative Commons 4 for original code
#  GPL3 for rewrite

import argparse
import sys
import time
from glob import glob
from pathlib import Path
from typing import Any, Tuple, List, Dict

import yaml

from oewn_core.wordnet import WordnetModel, Entry, Sense, Synset, PartOfSpeech, Example, Pronunciation, VerbFrame


def load_verbframes(home: str) -> List[VerbFrame]:
    """
    Load verb frames from YAML
    :param home: home dir for YAML frame.yaml file
    :return: list of verb frames
    """
    with open(f'{home}/frames.yaml', encoding='utf-8') as inp:
        y: Dict[str, Any] = yaml.load(inp, Loader=yaml.CLoader)
        return [VerbFrame(k, v) for k, v in y.items()]


def load_entries(home: str) -> Tuple[List[Entry], Dict[str, Sense], Dict[Tuple[str, str], Entry]]:
    """
    Load entries from YAML
    :param home: home dir for YAML entries-*.yaml file
    :return: list of entries, sense resolver, member resolver
    """
    sense_resolver: Dict[str, Sense] = {}
    member_resolver: Dict[Tuple[str, str], Entry] = {}
    entries: List[Entry] = []
    for f in glob(f'{home}/entries-*.yaml'):
        with open(f, encoding='utf-8') as inp:
            y: Dict[str, Any] = yaml.load(inp, Loader=yaml.CLoader)
            for lemma, poses_discriminants in y.items():
                for pos_discriminant, entry_y in poses_discriminants.items():
                    pos = PartOfSpeech(pos_discriminant[:1]).value
                    discriminant = pos_discriminant[2:] if len(pos_discriminant) > 2 else None
                    entry = Entry(lemma, pos, discriminant)
                    if 'form' in entry_y:
                        entry.forms = entry_y['form']
                    if 'pronunciation' in entry_y:
                        entry.pronunciations = [Pronunciation(p['value'], p.get('variety')) for p in entry_y['pronunciation']]
                    for n, sense_y in enumerate(entry_y['sense']):
                        sense = load_sense(sense_y, entry)
                        entry.senses.append(sense)
                        sense_resolver[sense.id] = sense
                        member_resolver[(lemma, sense.synsetid)] = entry
                    entries.append(entry)
    return entries, sense_resolver, member_resolver


def load_synsets(home: str) -> Tuple[List[Synset], Dict[str, Synset]]:
    """
    Load synsets from YAML
    :param home: home dir for YAML (noun|verb|adj|adv))-*.yaml file
    :return: list of synsets, synset resolver
    """
    resolver: Dict[str, Synset] = {}
    synsets: List[Synset] = []
    noun_files = glob(f'{home}/noun*.yaml')
    verb_files = glob(f'{home}/verb*.yaml')
    adj_files = glob(f'{home}/adj*.yaml')
    adv_files = glob(f'{home}/adv*.yaml')
    for f in noun_files + verb_files + adj_files + adv_files:
        lex_name = Path(f).stem
        with open(f, encoding='utf-8') as inp:
            y: Dict[str, Any] = yaml.load(inp, Loader=yaml.CLoader)
            for synsetid, synset_y in y.items():
                synset = load_synset(synset_y, synsetid, lex_name)
                synsets.append(synset)
                resolver[synsetid] = synset
    return synsets, resolver


def load_sense(y: Dict[str, Any], entry: Entry) -> Sense:
    """
    Load sense from YAML
    :param y: properties provided by PyYAML
    :param entry: wrapping entry
    :return: sense
    """
    s = Sense(y['id'], entry, y['synset'], y.get('adjposition'))
    if 'sent' in y:
        s.examples = y['sent']
    if 'subcat' in y:
        s.verbframeids = y['subcat']
    # relations
    sense_rel_types: List[str] = [str(t) for t in Sense.Relation.Type]
    other_rel_types: List[str] = [str(t) for t in Sense.Relation.OtherType]
    for rel, targets in y.items():
        if rel in sense_rel_types:
            for target in targets:
                s.relations.append(Sense.Relation(target, Sense.Relation.Type(rel).value))
        if rel in other_rel_types:
            for target in targets:
                s.relations.append(Sense.Relation(target, Sense.Relation.OtherType(rel).value, True))
    return s


def load_synset(y: Dict[str, Any], synsetid: str, lex_name: str) -> Synset:
    """
    Load synset from YAML
    :param y: properties provided by PyYAML
    :param synsetid: synset ID
    :param lex_name: lexical name, provided by file name's stem
    :return: synset
    """
    pos = PartOfSpeech(y['partOfSpeech']).value
    ss = Synset(synsetid, pos, y['members'], lex_name)
    for defn in y['definition']:
        ss.definitions.append(defn)
    for example in y.get('example', []):
        if isinstance(example, str):
            ss.examples.append(example)
        else:
            ss.examples.append(Example(example['text'], example['source']))
    for usage in y.get('usage', []):
        ss.usages.append(usage)
    ss.source = y.get('source')
    ss.wikidata = y.get('wikidata')
    ss.ili = y.get('ili', 'in')
    # relations
    synset_rel_types: List[str] = [str(t) for t in Synset.Relation.Type]
    for rel, targets in y.items():
        if rel in synset_rel_types:
            for target in targets:
                ss.relations.append(Synset.Relation(target, Synset.Relation.Type(rel).value))
    return ss


def load_core(home: str) -> WordnetModel:
    """
    Load synset from YAML
    :param home: home dir for YAML *.yaml file
    :return: unresolved, unextended model
    """
    wn = WordnetModel('oewn', 'Open English Wordnet', 'en',
                      'english-wordnet@googlegroups.com',
                      'https://creativecommons.org/licenses/by/4.0',
                      '2024',
                      'https://github.com/globalwordnet/english-wordnet')
    # lex entries
    wn.entries, wn.sense_resolver, wn.member_resolver = load_entries(home)

    # synsets
    wn.synsets, wn.synset_resolver = load_synsets(home)

    # frames
    wn.verbframes = load_verbframes(home)

    return wn


def load(home: str, extend: bool=True, resolve: bool=False, verbose: bool = False) -> WordnetModel:
    if verbose:
        print(f'loading from YAML in {home}')
    wn = load_core(home)
    if verbose:
        print(f'loaded {wn} from YAML in {home}')
    if extend:
        if verbose:
            print(f'extending relations')
            print(f'before extension: {wn.info_relations()}')
        wn.extend()
        if verbose:
            print(f'after extension:  {wn.info_relations()}')
            print(f'extended relations')
    if resolve:
        if verbose:
            print(f'resolving cross-references')
        wn.resolve()
        if verbose:
            print(f'resolved cross-references')
    if verbose:
        print(wn)
        print(wn.info())
        print(wn.info_relations())
    return wn


def main() -> WordnetModel:
    arg_parser = argparse.ArgumentParser(description="load from yaml")
    arg_parser.add_argument('in_dir', type=str, help='from-dir')
    args = arg_parser.parse_args()
    return load(args.in_dir)


if __name__ == '__main__':
    start_time = time.time()
    main()
    end_time = time.time()
    duration = end_time - start_time
    print(f"Loading took {duration:.6f} seconds", file=sys.stderr)



"""
WordNet model browse sample
Author: John McCrae <john@mccr.ae> for original code
Author: Bernard Bou <1313ou@gmail.com> for rewrite and revamp
"""
#  Copyright (c) 2024.
#  Creative Commons 4 for original code
#  GPL3 for rewrite

import argparse

from oewn_core.wordnet import WordnetModel


def browse(wn: WordnetModel) -> None:
    entity_resolver = wn.entry_resolver
    e = entity_resolver[('force', 'n', None)]
    print(f'{e}')
    for s in e.senses:
        ss = s.resolved_synset
        assert ss is not None
        print(f'\t{s} -> {ss.members} {ss.definitions}')
        for ssr in ss.relations:
            assert ssr.resolved_target
            print(f'\t\t{ssr} {ssr.resolved_target.members} {ssr.resolved_target.definitions}')
        for sr in s.relations:
            assert sr.resolved_target
            print(f'\t\t{sr} {sr.resolved_target}')


def main() -> None:
    def get_model() -> WordnetModel:
        if args.pickle:
            from oewn_core.deserialize import load
            return load(args.in_dir, file=args.pickled, resolve=True)
        else:
            from oewn_core.wordnet_fromyaml import load
            return load(args.in_dir, resolve=True)

    arg_parser = argparse.ArgumentParser(description="browse")
    arg_parser.add_argument('--pickle', action='store_true', default=False, help='use pickle')
    arg_parser.add_argument('in_dir', type=str, help='from-dir for yaml/pickle')
    arg_parser.add_argument('pickled', type=str, nargs='?', default='oewn.pickle', help='from-pickle')
    args = arg_parser.parse_args()

    wn = get_model()
    browse(wn)


if __name__ == '__main__':
    main()



#!/usr/bin/python3

"""
WordNet from-XML utilities
Uses a SAX parser

Author: John McCrae <john@mccr.ae> for original code
Author: Bernard Bou <1313ou@gmail.com> for rewrite and revamp
"""

#  Copyright (c) 2024.
#  Creative Commons 4 for original code
#  GPL3 for rewrite

import argparse
import codecs
import re
import sys
import time
from typing import Optional, List, Dict, Tuple
from xml.sax import parse
from xml.sax.handler import ContentHandler

from oewn_core.wordnet import WordnetModel, Entry, Sense, Synset, PartOfSpeech, Pronunciation, Example, VerbFrame
from oewn_xml.wordnet_xml import from_xml_synset_id, from_xml_sense_id


def make_synset_id(xml_synsetid: str) -> str:
    return from_xml_synset_id(xml_synsetid)


def make_sense_id(xml_senseid: str) -> str:
    return from_xml_sense_id(xml_senseid)


def make_member(xml_member: str, entry_resolver) -> str:
    e = entry_resolver[xml_member]
    return e.lemma


def make_members(xml_members: str, entry_resolver) -> List[str]:
    return [make_member(m, entry_resolver) for m in xml_members.split(' ')]


class SAXParser(ContentHandler):
    """
    SAX parser
    """

    def __init__(self) -> None:
        ContentHandler.__init__(self)

        # local data
        self.lexicon: Optional[WordnetModel] = None
        self.entry: Optional[Entry] = None
        self.sense: Optional[Sense] = None
        self.synset: Optional[Synset] = None
        self.defn: Optional[str] = None
        self.ili_defn: Optional[str] = None
        self.example: Optional[str] = None
        self.example_source: Optional[str] = None
        self.usage: Optional[str] = None
        self.pronunciation: Optional[str] = None
        self.pronunciation_variety: Optional[str] = None

        # top accumulators
        self.entries: List[Entry] = []
        self.synsets: List[Synset] = []
        self.verbframes: List[VerbFrame] = []
        # resolvers
        self.sense_resolver: Dict[str, Sense] = {}
        self.synset_resolver: Dict[str, Synset] = {}
        self.member_resolver: Dict[Tuple[str, str], Entry] = {}
        self.entry_resolver: Dict[str, Entry] = {}

    discriminant_pattern = r'\-\d+$'

    def startElement(self, name, attrs) -> None:
        if name == 'LexicalEntry':
            entryid = attrs.get('id')
            match = re.search(self.discriminant_pattern, entryid)
            d = match.group()[1:] if match else None
            self.entry = Entry(None, None, d)
            if entryid in self.entry_resolver:
                raise ValueError(f'Duplicate entry ID while parsing: {entryid}')
            self.entry_resolver[entryid] = self.entry
        elif name == 'Sense':
            senseid = from_xml_sense_id(attrs['id'])
            synsetid = make_synset_id(attrs['synset'])
            verbframes = attrs['subcat'].split(' ') if 'subcat' in attrs else None
            self.sense = Sense(senseid, self.entry, synsetid, attrs.get('adjposition'))
            self.sense.verbframeids = verbframes
        elif name == 'Synset':
            synsetid = make_synset_id(attrs['id'])
            members = make_members(attrs.get('members', ''), self.entry_resolver)
            pos = PartOfSpeech(attrs['partOfSpeech']).value
            self.synset = Synset(synsetid, pos, members, attrs.get('lexfile'))
            self.synset.ili = attrs['ili']
            self.synset.wikidata = attrs.get('dc:subject')
            self.synset.source = attrs.get('dc:source')
        elif name == 'Lemma':
            self.entry.lemma = attrs['writtenForm']
            self.entry.pos = attrs['partOfSpeech']
        elif name == 'Form':
            self.entry.forms.append(attrs['writtenForm'])
        elif name == 'Definition':
            self.defn = ''
        elif name == 'ILIDefinition':
            self.ili_defn = ''
        elif name == 'Example':
            self.example = ''
            self.example_source = attrs.get('dc:source')
        elif name == 'Usage':
            self.example = ''
        elif name == 'SynsetRelation':
            target = make_synset_id(attrs['target'])
            rtype = attrs['relType']
            self.synset.relations.append(Synset.Relation(target, Synset.Relation.Type(rtype).value))
        elif name == 'SenseRelation':
            target = make_sense_id(attrs['target'])
            rtype = attrs['relType']
            is_other = rtype == Sense.Relation.Type.OTHER.value
            rtype2 = Sense.Relation.OtherType(attrs['dc:type']).value if is_other else Sense.Relation.Type(rtype).value
            self.sense.relations.append(Sense.Relation(target, rtype2, is_other))
        elif name == 'SyntacticBehaviour':
            self.verbframes.append(VerbFrame(attrs['id'], attrs['subcategorizationFrame']))
        elif name == 'Pronunciation':
            self.pronunciation = ''
            self.pronunciation_variety = attrs.get('variety')
        elif name == 'Lexicon':
            self.lexicon = WordnetModel(
                attrs['id'],
                attrs['label'],
                attrs['language'],
                attrs['email'],
                attrs['license'],
                attrs['version'],
                attrs['url'])
        elif name == 'LexicalResource':
            pass
        else:
            raise ValueError(f'Unexpected Tag: {name}')

    def endElement(self, name) -> None:
        if name == 'LexicalEntry':
            self.entries.append(self.entry)
            self.entry = None
        elif name == 'Sense':
            assert self.entry
            self.entry.senses.append(self.sense)
            if self.sense.id in self.sense_resolver:
                raise ValueError(f'Duplicate sense ID while parsing: {self.sense.id}')
            self.sense_resolver[self.sense.id] = self.sense
            mk = (self.entry.lemma, self.sense.synsetid)
            if mk in self.member_resolver:
                raise ValueError(f'Duplicate member ID while parsing: {mk}')
            self.member_resolver[mk] = self.entry
            self.sense = None
        elif name == 'Synset':
            self.synsets.append(self.synset)
            if self.synset.id in self.member_resolver:
                raise ValueError(f'Duplicate synset ID while parsing: {self.synset.id}')
            self.synset_resolver[self.synset.id] = self.synset
            self.synset = None
        elif name == 'Definition':
            assert self.synset
            self.synset.definitions.append(self.defn)
            self.defn = None
        elif name == 'ILIDefinition':
            assert self.synset
            self.synset.ili_definition = self.ili_defn
            self.ili_defn = None
        elif name == 'Example':
            if self.synset:
                e = Example(self.example, self.example_source) if self.example_source else self.example
                self.synset.examples.append(e)
            elif self.sense:
                self.sense.examples.append(self.example)
            self.example = None
        elif name == 'Usage':
            assert self.synset
            self.synset.usages.append(self.usage)
            self.usage = None
        elif name == 'Pronunciation':
            assert self.entry
            self.entry.pronunciations.append(Pronunciation(self.pronunciation, self.pronunciation_variety))
            self.pronunciation = None

    def characters(self, content) -> None:
        if self.defn is not None:
            self.defn += content
        elif self.ili_defn is not None:
            self.ili_defn += content
        elif self.example is not None:
            self.example += content
        elif self.usage is not None:
            self.usage += content
        elif self.pronunciation is not None:
            self.pronunciation += content
        elif content.strip() == '':
            pass
        else:
            raise ValueError(f'Text content not expected: "{content}"')

    def get_parsed(self) -> WordnetModel:
        assert self.lexicon
        wn = self.lexicon
        wn.entries = self.entries
        wn.synsets = self.synsets
        wn.verbframes = self.verbframes
        wn.sense_resolver = self.sense_resolver
        wn.synset_resolver = self.synset_resolver
        wn.member_resolver = self.member_resolver
        return wn


def load_core(wordnet_file) -> WordnetModel:
    with codecs.open(wordnet_file, encoding='utf-8') as source:
        sax_parser = SAXParser()
        parse(source, sax_parser)
        return sax_parser.get_parsed()


def load(home: str, extend: bool = True, resolve: bool = False, verbose: bool = False) -> WordnetModel:
    if verbose:
        print(f'loading from XML in {home}')
    wn: WordnetModel = load_core(home)
    if verbose:
        print(f'loaded {wn} from XML in {home}')
    if extend:
        if verbose:
            print(f'extending relations')
            print(f'before extension: {wn.info_relations()}')
        wn.extend()
        if verbose:
            print(f'after extension:  {wn.info_relations()}')
            print(f'extended relations')
    if resolve:
        if verbose:
            print(f'resolving cross-references')
        wn.resolve()
        if verbose:
            print(f'resolved cross-references')
    if verbose:
        print(wn)
        print(wn.info())
        print(wn.info_relations())
    return wn


def main() -> WordnetModel:
    arg_parser = argparse.ArgumentParser(description="load from yaml")
    arg_parser.add_argument('in_dir', type=str, help='from-dir')
    args = arg_parser.parse_args()
    return load(args.in_dir)


if __name__ == '__main__':
    start_time = time.time()
    main()
    end_time = time.time()
    duration = end_time - start_time
    print(f"Loading took {duration:.6f} seconds", file=sys.stderr)
